# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from ydb._grpc.v4.protos import ydb_import_pb2 as protos_dot_ydb__import__pb2


class ImportServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.ImportFromS3 = channel.unary_unary(
                '/Ydb.Import.V1.ImportService/ImportFromS3',
                request_serializer=protos_dot_ydb__import__pb2.ImportFromS3Request.SerializeToString,
                response_deserializer=protos_dot_ydb__import__pb2.ImportFromS3Response.FromString,
                )
        self.ImportFromFs = channel.unary_unary(
                '/Ydb.Import.V1.ImportService/ImportFromFs',
                request_serializer=protos_dot_ydb__import__pb2.ImportFromFsRequest.SerializeToString,
                response_deserializer=protos_dot_ydb__import__pb2.ImportFromFsResponse.FromString,
                )
        self.ListObjectsInS3Export = channel.unary_unary(
                '/Ydb.Import.V1.ImportService/ListObjectsInS3Export',
                request_serializer=protos_dot_ydb__import__pb2.ListObjectsInS3ExportRequest.SerializeToString,
                response_deserializer=protos_dot_ydb__import__pb2.ListObjectsInS3ExportResponse.FromString,
                )
        self.ListObjectsInFsExport = channel.unary_unary(
                '/Ydb.Import.V1.ImportService/ListObjectsInFsExport',
                request_serializer=protos_dot_ydb__import__pb2.ListObjectsInFsExportRequest.SerializeToString,
                response_deserializer=protos_dot_ydb__import__pb2.ListObjectsInFsExportResponse.FromString,
                )
        self.ImportData = channel.unary_unary(
                '/Ydb.Import.V1.ImportService/ImportData',
                request_serializer=protos_dot_ydb__import__pb2.ImportDataRequest.SerializeToString,
                response_deserializer=protos_dot_ydb__import__pb2.ImportDataResponse.FromString,
                )


class ImportServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def ImportFromS3(self, request, context):
        """Imports data from S3.
        Method starts an asynchronous operation that can be cancelled while it is in progress.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ImportFromFs(self, request, context):
        """Imports data from file system.
        Method starts an asynchronous operation that can be cancelled while it is in progress.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ListObjectsInS3Export(self, request, context):
        """List objects from existing export stored in S3 bucket
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ListObjectsInFsExport(self, request, context):
        """List objects from existing export stored in FS
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ImportData(self, request, context):
        """Writes data to a table.
        Method accepts serialized data in the selected format and writes it non-transactionally.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_ImportServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'ImportFromS3': grpc.unary_unary_rpc_method_handler(
                    servicer.ImportFromS3,
                    request_deserializer=protos_dot_ydb__import__pb2.ImportFromS3Request.FromString,
                    response_serializer=protos_dot_ydb__import__pb2.ImportFromS3Response.SerializeToString,
            ),
            'ImportFromFs': grpc.unary_unary_rpc_method_handler(
                    servicer.ImportFromFs,
                    request_deserializer=protos_dot_ydb__import__pb2.ImportFromFsRequest.FromString,
                    response_serializer=protos_dot_ydb__import__pb2.ImportFromFsResponse.SerializeToString,
            ),
            'ListObjectsInS3Export': grpc.unary_unary_rpc_method_handler(
                    servicer.ListObjectsInS3Export,
                    request_deserializer=protos_dot_ydb__import__pb2.ListObjectsInS3ExportRequest.FromString,
                    response_serializer=protos_dot_ydb__import__pb2.ListObjectsInS3ExportResponse.SerializeToString,
            ),
            'ListObjectsInFsExport': grpc.unary_unary_rpc_method_handler(
                    servicer.ListObjectsInFsExport,
                    request_deserializer=protos_dot_ydb__import__pb2.ListObjectsInFsExportRequest.FromString,
                    response_serializer=protos_dot_ydb__import__pb2.ListObjectsInFsExportResponse.SerializeToString,
            ),
            'ImportData': grpc.unary_unary_rpc_method_handler(
                    servicer.ImportData,
                    request_deserializer=protos_dot_ydb__import__pb2.ImportDataRequest.FromString,
                    response_serializer=protos_dot_ydb__import__pb2.ImportDataResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'Ydb.Import.V1.ImportService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class ImportService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def ImportFromS3(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Ydb.Import.V1.ImportService/ImportFromS3',
            protos_dot_ydb__import__pb2.ImportFromS3Request.SerializeToString,
            protos_dot_ydb__import__pb2.ImportFromS3Response.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ImportFromFs(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Ydb.Import.V1.ImportService/ImportFromFs',
            protos_dot_ydb__import__pb2.ImportFromFsRequest.SerializeToString,
            protos_dot_ydb__import__pb2.ImportFromFsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ListObjectsInS3Export(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Ydb.Import.V1.ImportService/ListObjectsInS3Export',
            protos_dot_ydb__import__pb2.ListObjectsInS3ExportRequest.SerializeToString,
            protos_dot_ydb__import__pb2.ListObjectsInS3ExportResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ListObjectsInFsExport(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Ydb.Import.V1.ImportService/ListObjectsInFsExport',
            protos_dot_ydb__import__pb2.ListObjectsInFsExportRequest.SerializeToString,
            protos_dot_ydb__import__pb2.ListObjectsInFsExportResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ImportData(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Ydb.Import.V1.ImportService/ImportData',
            protos_dot_ydb__import__pb2.ImportDataRequest.SerializeToString,
            protos_dot_ydb__import__pb2.ImportDataResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



import asyncio
import logging
from typing import (
    Callable,
    Optional,
    List,
    Dict,
    Any,
    Union,
)

from .session import (
    QuerySession,
)
from ...retries import (
    RetrySettings,
    retry_operation_async,
)
from ...query.base import BaseQueryTxMode, QueryExplainResultFormat
from ...query.base import QueryClientSettings
from ... import convert
from ..._grpc.grpcwrapper import common_utils
from ..._grpc.grpcwrapper import ydb_query_public_types as _ydb_query_public

logger = logging.getLogger(__name__)


class QuerySessionPool:
    """QuerySessionPool is an object to simplify operations with sessions of Query Service."""

    def __init__(
        self,
        driver: common_utils.SupportedDriverType,
        size: int = 100,
        *,
        query_client_settings: Optional[QueryClientSettings] = None,
        loop: Optional[asyncio.AbstractEventLoop] = None,
    ):
        """
        :param driver: A driver instance
        :param size: Size of session pool
        :param query_client_settings: ydb.QueryClientSettings object to configure QueryService behavior
        """

        self._driver = driver
        self._size = size
        self._should_stop = asyncio.Event()
        self._queue = asyncio.Queue()
        self._current_size = 0
        self._waiters = 0
        self._loop = asyncio.get_running_loop() if loop is None else loop
        self._query_client_settings = query_client_settings

    async def _create_new_session(self):
        session = QuerySession(self._driver, settings=self._query_client_settings)
        await session.create()
        logger.debug(f"New session was created for pool. Session id: {session.session_id}")
        return session

    async def acquire(self) -> QuerySession:
        """Acquire a session from Session Pool.

        :return A QuerySession object.
        """

        if self._should_stop.is_set():
            logger.error("An attempt to take session from closed session pool.")
            raise RuntimeError("An attempt to take session from closed session pool.")

        session = None
        try:
            session = self._queue.get_nowait()
        except asyncio.QueueEmpty:
            pass

        if session is None and self._current_size == self._size:
            queue_get = asyncio.ensure_future(self._queue.get())
            task_stop = asyncio.ensure_future(asyncio.ensure_future(self._should_stop.wait()))
            try:
                done, _ = await asyncio.wait((queue_get, task_stop), return_when=asyncio.FIRST_COMPLETED)
            except asyncio.CancelledError:
                task_stop.cancel()
                cancelled = queue_get.cancel()
                if not cancelled and not queue_get.exception():
                    await self.release(queue_get.result())
                raise
            if task_stop in done:
                queue_get.cancel()
                raise RuntimeError("An attempt to take session from closed session pool.")

            task_stop.cancel()
            session = queue_get.result()

        if session is not None:
            if session.is_active:
                logger.debug(f"Acquired active session from queue: {session.session_id}")
                return session
            else:
                self._current_size -= 1
                logger.debug(f"Acquired dead session from queue: {session.session_id}")

        logger.debug(f"Session pool is not large enough: {self._current_size} < {self._size}, will create new one.")

        self._current_size += 1
        try:
            session = await self._create_new_session()
        except Exception as e:
            logger.error("Failed to create new session")
            self._current_size -= 1
            raise e

        return session

    async def release(self, session: QuerySession) -> None:
        """Release a session back to Session Pool."""

        self._queue.put_nowait(session)
        logger.debug("Session returned to queue: %s", session.session_id)

    def checkout(self) -> "SimpleQuerySessionCheckoutAsync":
        """Return a Session context manager, that acquires session on enter and releases session on exit."""

        return SimpleQuerySessionCheckoutAsync(self)

    async def retry_operation_async(
        self, callee: Callable, retry_settings: Optional[RetrySettings] = None, *args, **kwargs
    ):
        """Special interface to execute a bunch of commands with session in a safe, retriable way.

        :param callee: A function, that works with session.
        :param retry_settings: RetrySettings object.

        :return: Result sets or exception in case of execution errors.
        """

        retry_settings = RetrySettings() if retry_settings is None else retry_settings

        async def wrapped_callee():
            async with self.checkout() as session:
                return await callee(session, *args, **kwargs)

        return await retry_operation_async(wrapped_callee, retry_settings)

    async def retry_tx_async(
        self,
        callee: Callable,
        tx_mode: Optional[BaseQueryTxMode] = None,
        retry_settings: Optional[RetrySettings] = None,
        *args,
        **kwargs,
    ):
        """Special interface to execute a bunch of commands with transaction in a safe, retriable way.

        :param callee: A function, that works with session.
        :param tx_mode: Transaction mode, which is a one from the following choices:
          1) QuerySerializableReadWrite() which is default mode;
          2) QueryOnlineReadOnly(allow_inconsistent_reads=False);
          3) QuerySnapshotReadOnly();
          4) QuerySnapshotReadWrite();
          5) QueryStaleReadOnly().
        :param retry_settings: RetrySettings object.

        :return: Result sets or exception in case of execution errors.
        """

        tx_mode = tx_mode if tx_mode else _ydb_query_public.QuerySerializableReadWrite()
        retry_settings = RetrySettings() if retry_settings is None else retry_settings

        async def wrapped_callee():
            async with self.checkout() as session:
                async with session.transaction(tx_mode=tx_mode) as tx:
                    if tx_mode.name in ["serializable_read_write", "snapshot_read_only"]:
                        await tx.begin()
                    result = await callee(tx, *args, **kwargs)
                    await tx.commit()
                return result

        return await retry_operation_async(wrapped_callee, retry_settings)

    async def execute_with_retries(
        self,
        query: str,
        parameters: Optional[dict] = None,
        retry_settings: Optional[RetrySettings] = None,
        *args,
        **kwargs,
    ) -> List[convert.ResultSet]:
        """Special interface to execute a one-shot queries in a safe, retriable way.
        Note: this method loads all data from stream before return, do not use this
        method with huge read queries.

        :param query: A query, yql or sql text.
        :param parameters: dict with parameters and YDB types;
        :param retry_settings: RetrySettings object.

        :return: Result sets or exception in case of execution errors.
        """

        retry_settings = RetrySettings() if retry_settings is None else retry_settings

        async def wrapped_callee():
            async with self.checkout() as session:
                it = await session.execute(query, parameters, *args, **kwargs)
                return [result_set async for result_set in it]

        return await retry_operation_async(wrapped_callee, retry_settings)

    async def explain_with_retries(
        self,
        query: str,
        parameters: Optional[dict] = None,
        *,
        result_format: QueryExplainResultFormat = QueryExplainResultFormat.STR,
        retry_settings: Optional[RetrySettings] = None,
    ) -> Union[str, Dict[str, Any]]:
        """
        Explain a query in retriable way. No real query execution will happen.

        :param query: A query, yql or sql text.
        :param parameters: dict with parameters and YDB types;
        :param result_format: Return format: string or dict.
        :param retry_settings: RetrySettings object.
        :return: Parsed query plan.
        """

        async def callee(session: QuerySession):
            return await session.explain(query, parameters, result_format=result_format)

        return await self.retry_operation_async(callee, retry_settings)

    async def stop(self):
        self._should_stop.set()

        tasks = []
        while True:
            try:
                session = self._queue.get_nowait()
                tasks.append(session.delete())
            except asyncio.QueueEmpty:
                break

        await asyncio.gather(*tasks)

        logger.debug("All session were deleted.")

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.stop()


class SimpleQuerySessionCheckoutAsync:
    def __init__(self, pool: QuerySessionPool):
        self._pool = pool
        self._session = None

    async def __aenter__(self) -> QuerySession:
        self._session = await self._pool.acquire()
        return self._session

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self._pool.release(self._session)



# -*- coding: utf-8 -*-
# flake8: noqa
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: draft/ydb_dynamic_config_v1.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ydb._grpc.v3.draft.protos import ydb_dynamic_config_pb2 as draft_dot_protos_dot_ydb__dynamic__config__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='draft/ydb_dynamic_config_v1.proto',
  package='Ydb.DynamicConfig.V1',
  syntax='proto3',
  serialized_options=b'\n%tech.ydb.proto.draft.dynamicconfig.v1ZBgithub.com/ydb-platform/ydb-go-genproto/draft/Ydb_DynamicConfig_V1\370\001\001',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n!draft/ydb_dynamic_config_v1.proto\x12\x14Ydb.DynamicConfig.V1\x1a%draft/protos/ydb_dynamic_config.proto2\x81\x08\n\x14\x44ynamicConfigService\x12V\n\tSetConfig\x12#.Ydb.DynamicConfig.SetConfigRequest\x1a$.Ydb.DynamicConfig.SetConfigResponse\x12\x62\n\rReplaceConfig\x12\'.Ydb.DynamicConfig.ReplaceConfigRequest\x1a(.Ydb.DynamicConfig.ReplaceConfigResponse\x12\\\n\x0bGetMetadata\x12%.Ydb.DynamicConfig.GetMetadataRequest\x1a&.Ydb.DynamicConfig.GetMetadataResponse\x12V\n\tGetConfig\x12#.Ydb.DynamicConfig.GetConfigRequest\x1a$.Ydb.DynamicConfig.GetConfigResponse\x12Y\n\nDropConfig\x12$.Ydb.DynamicConfig.DropConfigRequest\x1a%.Ydb.DynamicConfig.DropConfigResponse\x12n\n\x11\x41\x64\x64VolatileConfig\x12+.Ydb.DynamicConfig.AddVolatileConfigRequest\x1a,.Ydb.DynamicConfig.AddVolatileConfigResponse\x12w\n\x14RemoveVolatileConfig\x12..Ydb.DynamicConfig.RemoveVolatileConfigRequest\x1a/.Ydb.DynamicConfig.RemoveVolatileConfigResponse\x12\x62\n\rGetNodeLabels\x12\'.Ydb.DynamicConfig.GetNodeLabelsRequest\x1a(.Ydb.DynamicConfig.GetNodeLabelsResponse\x12\x62\n\rResolveConfig\x12\'.Ydb.DynamicConfig.ResolveConfigRequest\x1a(.Ydb.DynamicConfig.ResolveConfigResponse\x12k\n\x10ResolveAllConfig\x12*.Ydb.DynamicConfig.ResolveAllConfigRequest\x1a+.Ydb.DynamicConfig.ResolveAllConfigResponseBn\n%tech.ydb.proto.draft.dynamicconfig.v1ZBgithub.com/ydb-platform/ydb-go-genproto/draft/Ydb_DynamicConfig_V1\xf8\x01\x01\x62\x06proto3'
  ,
  dependencies=[draft_dot_protos_dot_ydb__dynamic__config__pb2.DESCRIPTOR,])



_sym_db.RegisterFileDescriptor(DESCRIPTOR)


DESCRIPTOR._options = None

_DYNAMICCONFIGSERVICE = _descriptor.ServiceDescriptor(
  name='DynamicConfigService',
  full_name='Ydb.DynamicConfig.V1.DynamicConfigService',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=99,
  serialized_end=1124,
  methods=[
  _descriptor.MethodDescriptor(
    name='SetConfig',
    full_name='Ydb.DynamicConfig.V1.DynamicConfigService.SetConfig',
    index=0,
    containing_service=None,
    input_type=draft_dot_protos_dot_ydb__dynamic__config__pb2._SETCONFIGREQUEST,
    output_type=draft_dot_protos_dot_ydb__dynamic__config__pb2._SETCONFIGRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ReplaceConfig',
    full_name='Ydb.DynamicConfig.V1.DynamicConfigService.ReplaceConfig',
    index=1,
    containing_service=None,
    input_type=draft_dot_protos_dot_ydb__dynamic__config__pb2._REPLACECONFIGREQUEST,
    output_type=draft_dot_protos_dot_ydb__dynamic__config__pb2._REPLACECONFIGRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetMetadata',
    full_name='Ydb.DynamicConfig.V1.DynamicConfigService.GetMetadata',
    index=2,
    containing_service=None,
    input_type=draft_dot_protos_dot_ydb__dynamic__config__pb2._GETMETADATAREQUEST,
    output_type=draft_dot_protos_dot_ydb__dynamic__config__pb2._GETMETADATARESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetConfig',
    full_name='Ydb.DynamicConfig.V1.DynamicConfigService.GetConfig',
    index=3,
    containing_service=None,
    input_type=draft_dot_protos_dot_ydb__dynamic__config__pb2._GETCONFIGREQUEST,
    output_type=draft_dot_protos_dot_ydb__dynamic__config__pb2._GETCONFIGRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='DropConfig',
    full_name='Ydb.DynamicConfig.V1.DynamicConfigService.DropConfig',
    index=4,
    containing_service=None,
    input_type=draft_dot_protos_dot_ydb__dynamic__config__pb2._DROPCONFIGREQUEST,
    output_type=draft_dot_protos_dot_ydb__dynamic__config__pb2._DROPCONFIGRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='AddVolatileConfig',
    full_name='Ydb.DynamicConfig.V1.DynamicConfigService.AddVolatileConfig',
    index=5,
    containing_service=None,
    input_type=draft_dot_protos_dot_ydb__dynamic__config__pb2._ADDVOLATILECONFIGREQUEST,
    output_type=draft_dot_protos_dot_ydb__dynamic__config__pb2._ADDVOLATILECONFIGRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='RemoveVolatileConfig',
    full_name='Ydb.DynamicConfig.V1.DynamicConfigService.RemoveVolatileConfig',
    index=6,
    containing_service=None,
    input_type=draft_dot_protos_dot_ydb__dynamic__config__pb2._REMOVEVOLATILECONFIGREQUEST,
    output_type=draft_dot_protos_dot_ydb__dynamic__config__pb2._REMOVEVOLATILECONFIGRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetNodeLabels',
    full_name='Ydb.DynamicConfig.V1.DynamicConfigService.GetNodeLabels',
    index=7,
    containing_service=None,
    input_type=draft_dot_protos_dot_ydb__dynamic__config__pb2._GETNODELABELSREQUEST,
    output_type=draft_dot_protos_dot_ydb__dynamic__config__pb2._GETNODELABELSRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ResolveConfig',
    full_name='Ydb.DynamicConfig.V1.DynamicConfigService.ResolveConfig',
    index=8,
    containing_service=None,
    input_type=draft_dot_protos_dot_ydb__dynamic__config__pb2._RESOLVECONFIGREQUEST,
    output_type=draft_dot_protos_dot_ydb__dynamic__config__pb2._RESOLVECONFIGRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ResolveAllConfig',
    full_name='Ydb.DynamicConfig.V1.DynamicConfigService.ResolveAllConfig',
    index=9,
    containing_service=None,
    input_type=draft_dot_protos_dot_ydb__dynamic__config__pb2._RESOLVEALLCONFIGREQUEST,
    output_type=draft_dot_protos_dot_ydb__dynamic__config__pb2._RESOLVEALLCONFIGRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_DYNAMICCONFIGSERVICE)

DESCRIPTOR.services_by_name['DynamicConfigService'] = _DYNAMICCONFIGSERVICE

# @@protoc_insertion_point(module_scope)



import pytest
import time
import grpc
import threading
import tempfile
import os
import json
import asyncio
from unittest.mock import patch, AsyncMock, MagicMock

import tests.auth.test_credentials
import tests.oauth2_token_exchange
import tests.oauth2_token_exchange.test_token_exchange
import ydb.aio.iam
import ydb.aio.oauth2_token_exchange
import ydb.oauth2_token_exchange.token_source


class ServiceAccountCredentialsForTest(ydb.aio.iam.ServiceAccountCredentials):
    def _channel_factory(self):
        return grpc.aio.insecure_channel(self._iam_endpoint)

    def get_expire_time(self):
        return self._expires_in - time.time()


class Oauth2TokenExchangeCredentialsForTest(ydb.aio.oauth2_token_exchange.Oauth2TokenExchangeCredentials):
    def get_expire_time(self):
        return self._expires_in - time.time()


@pytest.mark.asyncio
async def test_yandex_service_account_credentials():
    server = tests.auth.test_credentials.IamTokenServiceTestServer()
    credentials = ServiceAccountCredentialsForTest(
        tests.auth.test_credentials.SERVICE_ACCOUNT_ID,
        tests.auth.test_credentials.ACCESS_KEY_ID,
        tests.auth.test_credentials.PRIVATE_KEY,
        server.get_endpoint(),
    )
    t = await credentials.get_auth_token()
    assert t == "test_token"
    assert credentials.get_expire_time() <= 42

    server.stop()


@pytest.mark.asyncio
async def test_oauth2_token_exchange_credentials():
    server = tests.oauth2_token_exchange.test_token_exchange.Oauth2TokenExchangeServiceForTest(40124)

    def serve(s):
        s.handle_request()

    serve_thread = threading.Thread(target=serve, args=(server,))
    serve_thread.start()

    credentials = Oauth2TokenExchangeCredentialsForTest(
        server.endpoint(),
        ydb.oauth2_token_exchange.token_source.FixedTokenSource("test_src_token", "test_token_type"),
        audience=["a1", "a2"],
        scope=["s1", "s2"],
    )
    t = (await credentials.auth_metadata())[0][1]
    assert t == "Bearer test_dst_token"
    assert credentials.get_expire_time() <= 42

    serve_thread.join()


@pytest.mark.asyncio
async def test_oauth2_token_exchange_credentials_file():
    server = tests.oauth2_token_exchange.test_token_exchange.Oauth2TokenExchangeServiceForTest(40124)

    def serve(s):
        s.handle_request()

    serve_thread = threading.Thread(target=serve, args=(server,))
    serve_thread.start()

    cfg = {
        "subject-credentials": {
            "type": "FIXED",
            "token": "test_src_token",
            "token-type": "test_token_type",
        },
        "aud": [
            "a1",
            "a2",
        ],
        "scope": [
            "s1",
            "s2",
        ],
    }

    temp_cfg_file = tempfile.NamedTemporaryFile(delete=False)
    cfg_file_name = temp_cfg_file.name

    try:
        temp_cfg_file.write(json.dumps(cfg, indent=4).encode("utf-8"))
        temp_cfg_file.close()

        credentials = Oauth2TokenExchangeCredentialsForTest.from_file(
            cfg_file=cfg_file_name, iam_endpoint=server.endpoint()
        )

        t = (await credentials.auth_metadata())[0][1]
        assert t == "Bearer test_dst_token"
        assert credentials.get_expire_time() <= 42

        serve_thread.join()
        os.remove(cfg_file_name)
    except Exception:
        os.remove(cfg_file_name)
        raise


@pytest.mark.asyncio
async def test_token_lazy_refresh():
    credentials = ServiceAccountCredentialsForTest(
        tests.auth.test_credentials.SERVICE_ACCOUNT_ID,
        tests.auth.test_credentials.ACCESS_KEY_ID,
        tests.auth.test_credentials.PRIVATE_KEY,
        "localhost:0",
    )

    credentials._tp.submit = MagicMock()

    mock_response = {"access_token": "token_v1", "expires_in": 3600}
    credentials._make_token_request = AsyncMock(return_value=mock_response)

    with patch("time.time") as mock_time:
        mock_time.return_value = 1000

        token1 = await credentials.token()
        assert token1 == "token_v1"
        assert credentials._make_token_request.call_count == 1

        token2 = await credentials.token()
        assert token2 == "token_v1"
        assert credentials._make_token_request.call_count == 1

        mock_time.return_value = 1000 + 3600 - 30 + 1
        credentials._make_token_request.return_value = {"access_token": "token_v2", "expires_in": 3600}

        token3 = await credentials.token()
        assert token3 == "token_v2"
        assert credentials._make_token_request.call_count == 2


@pytest.mark.asyncio
async def test_token_double_check_locking():
    credentials = ServiceAccountCredentialsForTest(
        tests.auth.test_credentials.SERVICE_ACCOUNT_ID,
        tests.auth.test_credentials.ACCESS_KEY_ID,
        tests.auth.test_credentials.PRIVATE_KEY,
        "localhost:0",
    )

    credentials._tp.submit = MagicMock()

    call_count = 0

    async def mock_make_request():
        nonlocal call_count
        call_count += 1
        await asyncio.sleep(0.01)
        return {"access_token": f"token_v{call_count}", "expires_in": 3600}

    credentials._make_token_request = mock_make_request

    with patch("time.time") as mock_time:
        mock_time.return_value = 1000

        tasks = [credentials.token() for _ in range(10)]
        results = await asyncio.gather(*tasks)

        assert len(set(results)) == 1
        assert call_count == 1


@pytest.mark.asyncio
async def test_token_expiration_calculation():
    credentials = ServiceAccountCredentialsForTest(
        tests.auth.test_credentials.SERVICE_ACCOUNT_ID,
        tests.auth.test_credentials.ACCESS_KEY_ID,
        tests.auth.test_credentials.PRIVATE_KEY,
        "localhost:0",
    )

    credentials._tp.submit = MagicMock()

    with patch("time.time") as mock_time:
        mock_time.return_value = 1000

        credentials._make_token_request = AsyncMock(return_value={"access_token": "token", "expires_in": 3600})

        await credentials.token()

        expected_expires = 1000 + 3600 - 30
        assert credentials._expires_in == expected_expires


@pytest.mark.asyncio
async def test_token_refresh_error_handling():
    credentials = ServiceAccountCredentialsForTest(
        tests.auth.test_credentials.SERVICE_ACCOUNT_ID,
        tests.auth.test_credentials.ACCESS_KEY_ID,
        tests.auth.test_credentials.PRIVATE_KEY,
        "localhost:0",
    )

    credentials._tp.submit = MagicMock()

    credentials._make_token_request = AsyncMock(side_effect=Exception("Network error"))

    with pytest.raises(Exception) as exc_info:
        await credentials.token()

    assert "Network error" in str(exc_info.value)
    assert credentials.last_error == "Network error"


@pytest.mark.asyncio
async def test_hybrid_background_and_sync_refresh():
    credentials = ServiceAccountCredentialsForTest(
        tests.auth.test_credentials.SERVICE_ACCOUNT_ID,
        tests.auth.test_credentials.ACCESS_KEY_ID,
        tests.auth.test_credentials.PRIVATE_KEY,
        "localhost:0",
    )

    call_count = 0
    background_calls = []

    async def mock_make_request():
        nonlocal call_count
        call_count += 1
        return {"access_token": f"token_v{call_count}", "expires_in": 3600}

    def mock_submit(callback):
        background_calls.append(callback)

    credentials._make_token_request = mock_make_request
    credentials._tp.submit = mock_submit

    with patch("time.time") as mock_time:
        mock_time.return_value = 1000

        token1 = await credentials.token()
        assert token1 == "token_v1"
        assert call_count == 1
        assert len(background_calls) == 0

        mock_time.return_value = 1000 + min(1800, 3600 / 10) + 1
        token2 = await credentials.token()
        assert token2 == "token_v1"
        assert call_count == 1
        assert len(background_calls) == 1

        mock_time.return_value = 1000 + 3600 - 30 + 1
        token3 = await credentials.token()
        assert token3 == "token_v2"
        assert call_count == 2



from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from collections.abc import Mapping as _Mapping
from typing import ClassVar as _ClassVar, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class ArrowBatchSettings(_message.Message):
    __slots__ = ("schema",)
    SCHEMA_FIELD_NUMBER: _ClassVar[int]
    schema: bytes
    def __init__(self, schema: _Optional[bytes] = ...) -> None: ...

class CsvSettings(_message.Message):
    __slots__ = ("skip_rows", "delimiter", "null_value", "header", "quoting")
    class Quoting(_message.Message):
        __slots__ = ("disabled", "quote_char", "double_quote_disabled")
        DISABLED_FIELD_NUMBER: _ClassVar[int]
        QUOTE_CHAR_FIELD_NUMBER: _ClassVar[int]
        DOUBLE_QUOTE_DISABLED_FIELD_NUMBER: _ClassVar[int]
        disabled: bool
        quote_char: bytes
        double_quote_disabled: bool
        def __init__(self, disabled: bool = ..., quote_char: _Optional[bytes] = ..., double_quote_disabled: bool = ...) -> None: ...
    SKIP_ROWS_FIELD_NUMBER: _ClassVar[int]
    DELIMITER_FIELD_NUMBER: _ClassVar[int]
    NULL_VALUE_FIELD_NUMBER: _ClassVar[int]
    HEADER_FIELD_NUMBER: _ClassVar[int]
    QUOTING_FIELD_NUMBER: _ClassVar[int]
    skip_rows: int
    delimiter: bytes
    null_value: bytes
    header: bool
    quoting: CsvSettings.Quoting
    def __init__(self, skip_rows: _Optional[int] = ..., delimiter: _Optional[bytes] = ..., null_value: _Optional[bytes] = ..., header: bool = ..., quoting: _Optional[_Union[CsvSettings.Quoting, _Mapping]] = ...) -> None: ...

class ArrowFormatSettings(_message.Message):
    __slots__ = ("compression_codec",)
    class CompressionCodec(_message.Message):
        __slots__ = ("type", "level")
        class Type(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
            __slots__ = ()
            TYPE_UNSPECIFIED: _ClassVar[ArrowFormatSettings.CompressionCodec.Type]
            TYPE_NONE: _ClassVar[ArrowFormatSettings.CompressionCodec.Type]
            TYPE_ZSTD: _ClassVar[ArrowFormatSettings.CompressionCodec.Type]
            TYPE_LZ4_FRAME: _ClassVar[ArrowFormatSettings.CompressionCodec.Type]
        TYPE_UNSPECIFIED: ArrowFormatSettings.CompressionCodec.Type
        TYPE_NONE: ArrowFormatSettings.CompressionCodec.Type
        TYPE_ZSTD: ArrowFormatSettings.CompressionCodec.Type
        TYPE_LZ4_FRAME: ArrowFormatSettings.CompressionCodec.Type
        TYPE_FIELD_NUMBER: _ClassVar[int]
        LEVEL_FIELD_NUMBER: _ClassVar[int]
        type: ArrowFormatSettings.CompressionCodec.Type
        level: int
        def __init__(self, type: _Optional[_Union[ArrowFormatSettings.CompressionCodec.Type, str]] = ..., level: _Optional[int] = ...) -> None: ...
    COMPRESSION_CODEC_FIELD_NUMBER: _ClassVar[int]
    compression_codec: ArrowFormatSettings.CompressionCodec
    def __init__(self, compression_codec: _Optional[_Union[ArrowFormatSettings.CompressionCodec, _Mapping]] = ...) -> None: ...

class ArrowFormatMeta(_message.Message):
    __slots__ = ("schema",)
    SCHEMA_FIELD_NUMBER: _ClassVar[int]
    schema: bytes
    def __init__(self, schema: _Optional[bytes] = ...) -> None: ...



# -*- coding: utf-8 -*-
# flake8: noqa
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: protos/ydb_scheme.proto
# Protobuf Python Version: 6.31.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    31,
    1,
    '',
    'protos/ydb_scheme.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ydb._grpc.v6.protos import ydb_common_pb2 as protos_dot_ydb__common__pb2
from ydb._grpc.v6.protos import ydb_operation_pb2 as protos_dot_ydb__operation__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x17protos/ydb_scheme.proto\x12\nYdb.Scheme\x1a\x17protos/ydb_common.proto\x1a\x1aprotos/ydb_operation.proto\"_\n\x14MakeDirectoryRequest\x12\x39\n\x10operation_params\x18\x01 \x01(\x0b\x32\x1f.Ydb.Operations.OperationParams\x12\x0c\n\x04path\x18\x02 \x01(\t\"E\n\x15MakeDirectoryResponse\x12,\n\toperation\x18\x01 \x01(\x0b\x32\x19.Ydb.Operations.Operation\"a\n\x16RemoveDirectoryRequest\x12\x39\n\x10operation_params\x18\x01 \x01(\x0b\x32\x1f.Ydb.Operations.OperationParams\x12\x0c\n\x04path\x18\x02 \x01(\t\"G\n\x17RemoveDirectoryResponse\x12,\n\toperation\x18\x01 \x01(\x0b\x32\x19.Ydb.Operations.Operation\"_\n\x14ListDirectoryRequest\x12\x39\n\x10operation_params\x18\x01 \x01(\x0b\x32\x1f.Ydb.Operations.OperationParams\x12\x0c\n\x04path\x18\x02 \x01(\t\"E\n\x15ListDirectoryResponse\x12,\n\toperation\x18\x01 \x01(\x0b\x32\x19.Ydb.Operations.Operation\"8\n\x0bPermissions\x12\x0f\n\x07subject\x18\x01 \x01(\t\x12\x18\n\x10permission_names\x18\x02 \x03(\t\"\xc1\x04\n\x05\x45ntry\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x05owner\x18\x02 \x01(\t\x12$\n\x04type\x18\x05 \x01(\x0e\x32\x16.Ydb.Scheme.Entry.Type\x12\x36\n\x15\x65\x66\x66\x65\x63tive_permissions\x18\x06 \x03(\x0b\x32\x17.Ydb.Scheme.Permissions\x12,\n\x0bpermissions\x18\x07 \x03(\x0b\x32\x17.Ydb.Scheme.Permissions\x12\x12\n\nsize_bytes\x18\x08 \x01(\x04\x12)\n\ncreated_at\x18\t \x01(\x0b\x32\x15.Ydb.VirtualTimestamp\"\xcf\x02\n\x04Type\x12\x14\n\x10TYPE_UNSPECIFIED\x10\x00\x12\r\n\tDIRECTORY\x10\x01\x12\t\n\x05TABLE\x10\x02\x12\x14\n\x10PERS_QUEUE_GROUP\x10\x03\x12\x0c\n\x08\x44\x41TABASE\x10\x04\x12\x0f\n\x0bRTMR_VOLUME\x10\x05\x12\x16\n\x12\x42LOCK_STORE_VOLUME\x10\x06\x12\x15\n\x11\x43OORDINATION_NODE\x10\x07\x12\x10\n\x0c\x43OLUMN_STORE\x10\x0c\x12\x10\n\x0c\x43OLUMN_TABLE\x10\r\x12\x0c\n\x08SEQUENCE\x10\x0f\x12\x0f\n\x0bREPLICATION\x10\x10\x12\t\n\x05TOPIC\x10\x11\x12\x12\n\x0e\x45XTERNAL_TABLE\x10\x12\x12\x18\n\x14\x45XTERNAL_DATA_SOURCE\x10\x13\x12\x08\n\x04VIEW\x10\x14\x12\x11\n\rRESOURCE_POOL\x10\x15\x12\x0c\n\x08TRANSFER\x10\x17\x12\x0c\n\x08SYS_VIEW\x10\x18\"[\n\x13ListDirectoryResult\x12\x1f\n\x04self\x18\x01 \x01(\x0b\x32\x11.Ydb.Scheme.Entry\x12#\n\x08\x63hildren\x18\x02 \x03(\x0b\x32\x11.Ydb.Scheme.Entry\"^\n\x13\x44\x65scribePathRequest\x12\x39\n\x10operation_params\x18\x01 \x01(\x0b\x32\x1f.Ydb.Operations.OperationParams\x12\x0c\n\x04path\x18\x02 \x01(\t\"D\n\x14\x44\x65scribePathResponse\x12,\n\toperation\x18\x01 \x01(\x0b\x32\x19.Ydb.Operations.Operation\"5\n\x12\x44\x65scribePathResult\x12\x1f\n\x04self\x18\x01 \x01(\x0b\x32\x11.Ydb.Scheme.Entry\"\xb2\x01\n\x11PermissionsAction\x12(\n\x05grant\x18\x01 \x01(\x0b\x32\x17.Ydb.Scheme.PermissionsH\x00\x12)\n\x06revoke\x18\x02 \x01(\x0b\x32\x17.Ydb.Scheme.PermissionsH\x00\x12&\n\x03set\x18\x03 \x01(\x0b\x32\x17.Ydb.Scheme.PermissionsH\x00\x12\x16\n\x0c\x63hange_owner\x18\x04 \x01(\tH\x00\x42\x08\n\x06\x61\x63tion\"\xde\x01\n\x18ModifyPermissionsRequest\x12\x39\n\x10operation_params\x18\x01 \x01(\x0b\x32\x1f.Ydb.Operations.OperationParams\x12\x0c\n\x04path\x18\x02 \x01(\t\x12.\n\x07\x61\x63tions\x18\x03 \x03(\x0b\x32\x1d.Ydb.Scheme.PermissionsAction\x12\x19\n\x11\x63lear_permissions\x18\x04 \x01(\x08\x12\x1f\n\x15interrupt_inheritance\x18\x05 \x01(\x08H\x00\x42\r\n\x0binheritance\"I\n\x19ModifyPermissionsResponse\x12,\n\toperation\x18\x01 \x01(\x0b\x32\x19.Ydb.Operations.OperationBl\n\x15tech.ydb.proto.schemeB\x15SchemeOperationProtosZ9github.com/ydb-platform/ydb-go-genproto/protos/Ydb_Scheme\xf8\x01\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'protos.ydb_scheme_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\025tech.ydb.proto.schemeB\025SchemeOperationProtosZ9github.com/ydb-platform/ydb-go-genproto/protos/Ydb_Scheme\370\001\001'
  _globals['_MAKEDIRECTORYREQUEST']._serialized_start=92
  _globals['_MAKEDIRECTORYREQUEST']._serialized_end=187
  _globals['_MAKEDIRECTORYRESPONSE']._serialized_start=189
  _globals['_MAKEDIRECTORYRESPONSE']._serialized_end=258
  _globals['_REMOVEDIRECTORYREQUEST']._serialized_start=260
  _globals['_REMOVEDIRECTORYREQUEST']._serialized_end=357
  _globals['_REMOVEDIRECTORYRESPONSE']._serialized_start=359
  _globals['_REMOVEDIRECTORYRESPONSE']._serialized_end=430
  _globals['_LISTDIRECTORYREQUEST']._serialized_start=432
  _globals['_LISTDIRECTORYREQUEST']._serialized_end=527
  _globals['_LISTDIRECTORYRESPONSE']._serialized_start=529
  _globals['_LISTDIRECTORYRESPONSE']._serialized_end=598
  _globals['_PERMISSIONS']._serialized_start=600
  _globals['_PERMISSIONS']._serialized_end=656
  _globals['_ENTRY']._serialized_start=659
  _globals['_ENTRY']._serialized_end=1236
  _globals['_ENTRY_TYPE']._serialized_start=901
  _globals['_ENTRY_TYPE']._serialized_end=1236
  _globals['_LISTDIRECTORYRESULT']._serialized_start=1238
  _globals['_LISTDIRECTORYRESULT']._serialized_end=1329
  _globals['_DESCRIBEPATHREQUEST']._serialized_start=1331
  _globals['_DESCRIBEPATHREQUEST']._serialized_end=1425
  _globals['_DESCRIBEPATHRESPONSE']._serialized_start=1427
  _globals['_DESCRIBEPATHRESPONSE']._serialized_end=1495
  _globals['_DESCRIBEPATHRESULT']._serialized_start=1497
  _globals['_DESCRIBEPATHRESULT']._serialized_end=1550
  _globals['_PERMISSIONSACTION']._serialized_start=1553
  _globals['_PERMISSIONSACTION']._serialized_end=1731
  _globals['_MODIFYPERMISSIONSREQUEST']._serialized_start=1734
  _globals['_MODIFYPERMISSIONSREQUEST']._serialized_end=1956
  _globals['_MODIFYPERMISSIONSRESPONSE']._serialized_start=1958
  _globals['_MODIFYPERMISSIONSRESPONSE']._serialized_end=2031
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# flake8: noqa
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: ydb_auth_v1.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ydb._grpc.v3.protos import ydb_auth_pb2 as protos_dot_ydb__auth__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='ydb_auth_v1.proto',
  package='Ydb.Auth.V1',
  syntax='proto3',
  serialized_options=b'\n\026tech.ydb.proto.auth.v1Z3github.com/ydb-platform/ydb-go-genproto/Ydb_Auth_V1',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x11ydb_auth_v1.proto\x12\x0bYdb.Auth.V1\x1a\x15protos/ydb_auth.proto2G\n\x0b\x41uthService\x12\x38\n\x05Login\x12\x16.Ydb.Auth.LoginRequest\x1a\x17.Ydb.Auth.LoginResponseBM\n\x16tech.ydb.proto.auth.v1Z3github.com/ydb-platform/ydb-go-genproto/Ydb_Auth_V1b\x06proto3'
  ,
  dependencies=[protos_dot_ydb__auth__pb2.DESCRIPTOR,])



_sym_db.RegisterFileDescriptor(DESCRIPTOR)


DESCRIPTOR._options = None

_AUTHSERVICE = _descriptor.ServiceDescriptor(
  name='AuthService',
  full_name='Ydb.Auth.V1.AuthService',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=57,
  serialized_end=128,
  methods=[
  _descriptor.MethodDescriptor(
    name='Login',
    full_name='Ydb.Auth.V1.AuthService.Login',
    index=0,
    containing_service=None,
    input_type=protos_dot_ydb__auth__pb2._LOGINREQUEST,
    output_type=protos_dot_ydb__auth__pb2._LOGINRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_AUTHSERVICE)

DESCRIPTOR.services_by_name['AuthService'] = _AUTHSERVICE

# @@protoc_insertion_point(module_scope)



import uuid

import pytest
import ydb


query_template = "DECLARE $a as %s; SELECT $a AS value"


def test_select_implicit_int(pool: ydb.QuerySessionPool):
    query = query_template % "Int64"
    expected_value = 111
    res = pool.execute_with_retries(query, parameters={"$a": expected_value})
    actual_value = res[0].rows[0]["value"]
    assert expected_value == actual_value


def test_select_implicit_float(pool: ydb.QuerySessionPool):
    query = query_template % "Double"
    expected_value = 11.1
    res = pool.execute_with_retries(query, parameters={"$a": expected_value})
    actual_value = res[0].rows[0]["value"]
    assert expected_value == pytest.approx(actual_value)


def test_select_implicit_bool(pool: ydb.QuerySessionPool):
    query = query_template % "Bool"
    expected_value = False
    res = pool.execute_with_retries(query, parameters={"$a": expected_value})
    actual_value = res[0].rows[0]["value"]
    assert expected_value == actual_value


def test_select_implicit_str(pool: ydb.QuerySessionPool):
    query = query_template % "Utf8"
    expected_value = "text"
    res = pool.execute_with_retries(query, parameters={"$a": expected_value})
    actual_value = res[0].rows[0]["value"]
    assert expected_value == actual_value


def test_select_implicit_bytes(pool: ydb.QuerySessionPool):
    query = query_template % "String"
    expected_value = b"text"
    res = pool.execute_with_retries(query, parameters={"$a": expected_value})
    actual_value = res[0].rows[0]["value"]
    assert expected_value == actual_value


def test_select_implicit_list(pool: ydb.QuerySessionPool):
    query = query_template % "List<Int64>"
    expected_value = [1, 2, 3]
    res = pool.execute_with_retries(query, parameters={"$a": expected_value})
    actual_value = res[0].rows[0]["value"]
    assert expected_value == actual_value


def test_select_implicit_dict(pool: ydb.QuerySessionPool):
    query = query_template % "Dict<Utf8, Int64>"
    expected_value = {"a": 1, "b": 2}
    res = pool.execute_with_retries(query, parameters={"$a": expected_value})
    actual_value = res[0].rows[0]["value"]
    assert expected_value == actual_value


def test_select_implicit_list_nested(pool: ydb.QuerySessionPool):
    query = query_template % "List<Dict<Utf8, Int64>>"
    expected_value = [{"a": 1}, {"b": 2}]
    res = pool.execute_with_retries(query, parameters={"$a": expected_value})
    actual_value = res[0].rows[0]["value"]
    assert expected_value == actual_value


def test_select_implicit_dict_nested(pool: ydb.QuerySessionPool):
    query = query_template % "Dict<Utf8, List<Int64>>"
    expected_value = {"a": [1, 2, 3], "b": [4, 5]}
    res = pool.execute_with_retries(query, parameters={"$a": expected_value})
    actual_value = res[0].rows[0]["value"]
    assert expected_value == actual_value


def test_select_implicit_custom_type_raises(pool: ydb.QuerySessionPool):
    query = query_template % "Struct"

    class CustomClass:
        pass

    expected_value = CustomClass()
    with pytest.raises(ValueError):
        pool.execute_with_retries(query, parameters={"$a": expected_value})


def test_select_implicit_empty_list_raises(pool: ydb.QuerySessionPool):
    query = query_template % "List<Int64>"
    expected_value = []
    with pytest.raises(ValueError):
        pool.execute_with_retries(query, parameters={"$a": expected_value})


def test_select_implicit_empty_dict_raises(pool: ydb.QuerySessionPool):
    query = query_template % "Dict<Int64, Int64>"
    expected_value = {}
    with pytest.raises(ValueError):
        pool.execute_with_retries(query, parameters={"$a": expected_value})


def test_select_explicit_primitive(pool: ydb.QuerySessionPool):
    query = query_template % "Int64"
    expected_value = 111
    res = pool.execute_with_retries(query, parameters={"$a": (expected_value, ydb.PrimitiveType.Int64)})
    actual_value = res[0].rows[0]["value"]
    assert expected_value == actual_value


def test_select_explicit_list(pool: ydb.QuerySessionPool):
    query = query_template % "List<Int64>"
    expected_value = [1, 2, 3]
    type_ = ydb.ListType(ydb.PrimitiveType.Int64)
    res = pool.execute_with_retries(query, parameters={"$a": (expected_value, type_)})
    actual_value = res[0].rows[0]["value"]
    assert expected_value == actual_value


def test_select_explicit_dict(pool: ydb.QuerySessionPool):
    query = query_template % "Dict<Utf8, Utf8>"
    expected_value = {"key": "value"}
    type_ = ydb.DictType(ydb.PrimitiveType.Utf8, ydb.PrimitiveType.Utf8)
    res = pool.execute_with_retries(query, parameters={"$a": (expected_value, type_)})
    actual_value = res[0].rows[0]["value"]
    assert expected_value == actual_value


def test_select_explicit_empty_list_not_raises(pool: ydb.QuerySessionPool):
    query = query_template % "List<Int64>"
    expected_value = []
    type_ = ydb.ListType(ydb.PrimitiveType.Int64)
    res = pool.execute_with_retries(query, parameters={"$a": (expected_value, type_)})
    actual_value = res[0].rows[0]["value"]
    assert expected_value == actual_value


def test_select_explicit_empty_dict_not_raises(pool: ydb.QuerySessionPool):
    query = query_template % "Dict<Utf8, Utf8>"
    expected_value = {}
    type_ = ydb.DictType(ydb.PrimitiveType.Utf8, ydb.PrimitiveType.Utf8)
    res = pool.execute_with_retries(query, parameters={"$a": (expected_value, type_)})
    actual_value = res[0].rows[0]["value"]
    assert expected_value == actual_value


def test_select_typedvalue_full_primitive(pool: ydb.QuerySessionPool):
    query = query_template % "Int64"
    expected_value = 111
    typed_value = ydb.TypedValue(expected_value, ydb.PrimitiveType.Int64)
    res = pool.execute_with_retries(query, parameters={"$a": typed_value})
    actual_value = res[0].rows[0]["value"]
    assert expected_value == actual_value


def test_select_typedvalue_implicit_primitive(pool: ydb.QuerySessionPool):
    query = query_template % "Int64"
    expected_value = 111
    typed_value = ydb.TypedValue(expected_value)
    res = pool.execute_with_retries(query, parameters={"$a": typed_value})
    actual_value = res[0].rows[0]["value"]
    assert expected_value == actual_value


def test_select_typevalue_custom_type_raises(pool: ydb.QuerySessionPool):
    query = query_template % "Struct"

    class CustomClass:
        pass

    expected_value = CustomClass()
    typed_value = ydb.TypedValue(expected_value)
    with pytest.raises(ValueError):
        pool.execute_with_retries(query, parameters={"$a": typed_value})


def test_uuid_send(pool: ydb.QuerySessionPool):
    val = uuid.UUID("6E73B41C-4EDE-4D08-9CFB-B7462D9E498B")
    query = """
DECLARE $val AS UUID;

SELECT CAST($val AS Utf8) AS value
"""
    res = pool.execute_with_retries(query, parameters={"$val": ydb.TypedValue(val, ydb.PrimitiveType.UUID)})
    actual_value = res[0].rows[0]["value"]
    assert actual_value.upper() == str(val).upper()


def test_uuid_read(pool: ydb.QuerySessionPool):
    val = uuid.UUID("6E73B41C-4EDE-4D08-9CFB-B7462D9E498B")
    query = """
DECLARE $val AS Utf8;

SELECT CAST($val AS UUID) AS value
"""
    res = pool.execute_with_retries(query, parameters={"$val": ydb.TypedValue(str(val), ydb.PrimitiveType.Utf8)})
    actual_value = res[0].rows[0]["value"]
    assert actual_value.hex.upper() == val.hex.upper()



import pytest
import ydb

from datetime import date, datetime, timedelta, timezone
from decimal import Decimal
from uuid import uuid4


@pytest.mark.parametrize(
    "value,ydb_type",
    [
        (True, ydb.PrimitiveType.Bool),
        (-125, ydb.PrimitiveType.Int8),
        (None, ydb.OptionalType(ydb.PrimitiveType.Int8)),
        (-32766, ydb.PrimitiveType.Int16),
        (-1123, ydb.PrimitiveType.Int32),
        (-2157583648, ydb.PrimitiveType.Int64),
        (255, ydb.PrimitiveType.Uint8),
        (65534, ydb.PrimitiveType.Uint16),
        (5555, ydb.PrimitiveType.Uint32),
        (2157583649, ydb.PrimitiveType.Uint64),
        (3.1415, ydb.PrimitiveType.Double),
        (".31415926535e1", ydb.PrimitiveType.DyNumber),
        (Decimal("3.1415926535"), ydb.DecimalType(28, 10)),
        (b"Hello, YDB!", ydb.PrimitiveType.String),
        ("Hello, 🐍!", ydb.PrimitiveType.Utf8),
        ('{"foo": "bar"}', ydb.PrimitiveType.Json),
        (b'{"foo"="bar"}', ydb.PrimitiveType.Yson),
        ('{"foo":"bar"}', ydb.PrimitiveType.JsonDocument),
        (uuid4(), ydb.PrimitiveType.UUID),
        ([1, 2, 3], ydb.ListType(ydb.PrimitiveType.Int8)),
        ({1: None, 2: None, 3: None}, ydb.SetType(ydb.PrimitiveType.Int8)),
        ([b"a", b"b", b"c"], ydb.ListType(ydb.PrimitiveType.String)),
        ({"a": 1001, "b": 1002}, ydb.DictType(ydb.PrimitiveType.Utf8, ydb.PrimitiveType.Int32)),
        (
            ("a", 1001),
            ydb.TupleType().add_element(ydb.PrimitiveType.Utf8).add_element(ydb.PrimitiveType.Int32),
        ),
        (
            {"foo": True, "bar": None},
            ydb.StructType()
            .add_member("foo", ydb.OptionalType(ydb.PrimitiveType.Bool))
            .add_member("bar", ydb.OptionalType(ydb.PrimitiveType.Int32)),
        ),
        (100, ydb.PrimitiveType.Date),
        (100, ydb.PrimitiveType.Date32),
        (-100, ydb.PrimitiveType.Date32),
        (100, ydb.PrimitiveType.Datetime),
        (100, ydb.PrimitiveType.Datetime64),
        (-100, ydb.PrimitiveType.Datetime64),
        (-100, ydb.PrimitiveType.Interval),
        (-100, ydb.PrimitiveType.Interval64),
        (100, ydb.PrimitiveType.Timestamp),
        (100, ydb.PrimitiveType.Timestamp64),
        (-100, ydb.PrimitiveType.Timestamp64),
        (1511789040123456, ydb.PrimitiveType.Timestamp),
        (1511789040123456, ydb.PrimitiveType.Timestamp64),
        (-1511789040123456, ydb.PrimitiveType.Timestamp64),
    ],
)
def test_types(driver_sync: ydb.Driver, value, ydb_type):
    settings = (
        ydb.QueryClientSettings()
        .with_native_date_in_result_sets(False)
        .with_native_datetime_in_result_sets(False)
        .with_native_timestamp_in_result_sets(False)
        .with_native_interval_in_result_sets(False)
        .with_native_json_in_result_sets(False)
    )
    with ydb.QuerySessionPool(driver_sync, query_client_settings=settings) as pool:
        result = pool.execute_with_retries(
            f"DECLARE $param as {ydb_type}; SELECT $param as value",
            {"$param": (value, ydb_type)},
        )
        assert result[0].rows[0].value == value


test_td = timedelta(microseconds=-100)
test_now = datetime.utcnow()
test_old_date = datetime(1221, 1, 1, 0, 0)
test_today = test_now.date()
test_dt_today = datetime.today()
tz4h = timezone(timedelta(hours=4))


@pytest.mark.parametrize(
    "value,ydb_type,result_value",
    [
        # FIXME: TypeError: 'datetime.datetime' object cannot be interpreted as an integer
        # (test_dt_today, "Datetime", test_dt_today),
        (test_today, ydb.PrimitiveType.Date, test_today),
        (365, ydb.PrimitiveType.Date, date(1971, 1, 1)),
        (-365, ydb.PrimitiveType.Date32, date(1969, 1, 1)),
        (3600 * 24 * 365, ydb.PrimitiveType.Datetime, datetime(1971, 1, 1, 0, 0)),
        (3600 * 24 * 365 * (-1), ydb.PrimitiveType.Datetime64, datetime(1969, 1, 1, 0, 0)),
        (datetime(1970, 1, 1, 4, 0, tzinfo=tz4h), ydb.PrimitiveType.Timestamp, datetime(1970, 1, 1, 0, 0)),
        (test_td, ydb.PrimitiveType.Interval, test_td),
        (test_td, ydb.PrimitiveType.Interval64, test_td),
        (test_now, ydb.PrimitiveType.Timestamp, test_now),
        (test_old_date, ydb.PrimitiveType.Timestamp64, test_old_date),
        (
            1511789040123456,
            ydb.PrimitiveType.Timestamp,
            datetime.fromisoformat("2017-11-27 13:24:00.123456"),
        ),
        ('{"foo": "bar"}', ydb.PrimitiveType.Json, {"foo": "bar"}),
        ('{"foo": "bar"}', ydb.PrimitiveType.JsonDocument, {"foo": "bar"}),
    ],
)
def test_types_native(driver_sync, value, ydb_type, result_value):
    with ydb.QuerySessionPool(driver_sync) as pool:
        result = pool.execute_with_retries(
            f"DECLARE $param as {ydb_type}; SELECT $param as value",
            {"$param": (value, ydb_type)},
        )
        assert result[0].rows[0].value == result_value


@pytest.mark.parametrize(
    "value,ydb_type,str_repr,result_value",
    [
        (test_today, ydb.PrimitiveType.Date, str(test_today), test_today),
        (365, ydb.PrimitiveType.Date, "1971-01-01", date(1971, 1, 1)),
        (-365, ydb.PrimitiveType.Date32, "1969-01-01", date(1969, 1, 1)),
        (3600 * 24 * 365, ydb.PrimitiveType.Datetime, "1971-01-01T00:00:00Z", datetime(1971, 1, 1, 0, 0)),
        (3600 * 24 * 365 * (-1), ydb.PrimitiveType.Datetime64, "1969-01-01T00:00:00Z", datetime(1969, 1, 1, 0, 0)),
        (
            datetime(1970, 1, 1, 4, 0, tzinfo=tz4h),
            ydb.PrimitiveType.Timestamp,
            "1970-01-01T00:00:00Z",
            datetime(1970, 1, 1, 0, 0),
        ),
        (test_td, ydb.PrimitiveType.Interval, "-PT0.0001S", test_td),
        (test_td, ydb.PrimitiveType.Interval64, "-PT0.0001S", test_td),
        (test_old_date, ydb.PrimitiveType.Timestamp64, "1221-01-01T00:00:00Z", test_old_date),
    ],
)
def test_type_str_repr(driver_sync, value, ydb_type, str_repr, result_value):
    with ydb.QuerySessionPool(driver_sync) as pool:
        result = pool.execute_with_retries(
            f"DECLARE $param as {ydb_type}; SELECT CAST($param as Utf8) as value",
            {"$param": (value, ydb_type)},
        )
        assert result[0].rows[0].value == str_repr

        result = pool.execute_with_retries(
            f"DECLARE $param as Utf8; SELECT CAST($param as {ydb_type}) as value",
            {"$param": (str_repr, ydb.PrimitiveType.Utf8)},
        )
        assert result[0].rows[0].value == result_value



# -*- coding: utf-8 -*-
import abc
import typing

from . import tracing, issues, connection
from . import settings as settings_impl
from concurrent import futures
import threading
import logging
import time

# Workaround for good IDE and universal for runtime
if typing.TYPE_CHECKING:
    from ._grpc.v4.protos import ydb_auth_pb2
    from ._grpc.v4 import ydb_auth_v1_pb2_grpc
else:
    from ._grpc.common.protos import ydb_auth_pb2
    from ._grpc.common import ydb_auth_v1_pb2_grpc


YDB_AUTH_TICKET_HEADER = "x-ydb-auth-ticket"
logger = logging.getLogger(__name__)


class AtMostOneExecution(object):
    def __init__(self):
        self._can_schedule = True
        self._lock = threading.Lock()
        self._tp = futures.ThreadPoolExecutor(1)

    def wrapped_execution(self, callback):
        try:
            callback()
        except Exception:
            pass

        finally:
            self.cleanup()

    def submit(self, callback):
        with self._lock:
            if self._can_schedule:
                self._tp.submit(self.wrapped_execution, callback)
                self._can_schedule = False

    def cleanup(self):
        with self._lock:
            self._can_schedule = True


class AbstractCredentials(abc.ABC):
    """
    An abstract class that provides auth metadata
    """


class Credentials(abc.ABC):
    def __init__(self, tracer=None):
        self.tracer = tracer if tracer is not None else tracing.Tracer(None)

    @abc.abstractmethod
    def auth_metadata(self):
        """
        :return: An iterable with auth metadata
        """
        pass

    def get_auth_token(self) -> str:
        for header, token in self.auth_metadata():
            if header == YDB_AUTH_TICKET_HEADER:
                return token
        return ""

    def _update_driver_config(self, driver_config):
        pass


class AbstractExpiringTokenCredentials(Credentials):
    def __init__(self, tracer=None):
        super(AbstractExpiringTokenCredentials, self).__init__(tracer)
        self._refresh_in = 0
        self._expires_in = 0
        self._cached_token = None
        self._token_lock = threading.Lock()
        self.logger = logger.getChild(self.__class__.__name__)
        self.last_error = None
        self.extra_error_message = ""
        self._hour = 60 * 60
        self._tp = AtMostOneExecution()
        self._time_shift_protection_seconds = 30

    @abc.abstractmethod
    def _make_token_request(self):
        pass

    def _is_token_valid(self):
        return self._cached_token is not None and time.time() <= self._expires_in

    def _should_refresh(self):
        return time.time() >= self._refresh_in

    def _update_token_info(self, token_response, current_time):
        self._refresh_in = current_time + min(self._hour / 2, token_response["expires_in"] / 10)
        self._expires_in = current_time + token_response["expires_in"] - self._time_shift_protection_seconds
        self._cached_token = token_response["access_token"]

    def _refresh_token(self, should_raise=False):
        current_time = time.time()

        try:
            self.logger.debug("Refreshing token, current_time: %s, expires_in: %s", current_time, self._expires_in)

            token_response = self._make_token_request()
            self._update_token_info(token_response, current_time)

            self.logger.info("Token refreshed successfully, expires_in: %s", self._expires_in)
            self.last_error = None

        except Exception as e:
            self.last_error = str(e)
            self.logger.error("Failed to refresh token: %s", e)
            if should_raise:
                raise issues.ConnectionError(
                    "%s: %s.\n%s" % (self.__class__.__name__, self.last_error, self.extra_error_message)
                )

    @property
    @tracing.with_trace()
    def token(self):
        if self._is_token_valid():
            if self._should_refresh():
                tracing.trace(self.tracer, {"refresh": True})
                self._tp.submit(self._refresh_token)

            tracing.trace(self.tracer, {"consumed": True})
            return self._cached_token

        with self._token_lock:
            if self._is_token_valid():
                tracing.trace(self.tracer, {"consumed": True})
                return self._cached_token

            tracing.trace(self.tracer, {"refresh": True})
            self._refresh_token(should_raise=True)

        tracing.trace(self.tracer, {"consumed": True})
        return self._cached_token

    def auth_metadata(self):
        return [(YDB_AUTH_TICKET_HEADER, self.token)]


def _wrap_static_credentials_response(rpc_state, response):
    issues._process_response(response.operation)
    result = ydb_auth_pb2.LoginResult()
    response.operation.result.Unpack(result)
    return result


class StaticCredentials(AbstractExpiringTokenCredentials):
    def __init__(self, driver_config, user, password="", tracer=None):
        super(StaticCredentials, self).__init__(tracer)

        from .driver import DriverConfig

        if driver_config is not None:
            self.driver_config = DriverConfig(
                endpoint=driver_config.endpoint,
                database=driver_config.database,
                root_certificates=driver_config.root_certificates,
            )
        self.user = user
        self.password = password
        self.request_timeout = 10

    @classmethod
    def from_user_password(cls, user: str, password: str, tracer=None):
        return cls(None, user, password, tracer)

    def _make_token_request(self):
        conn = connection.Connection.ready_factory(self.driver_config.endpoint, self.driver_config)
        assert conn is not None, "Failed to establish connection in to %s" % self.driver_config.endpoint
        try:
            result = conn(
                ydb_auth_pb2.LoginRequest(user=self.user, password=self.password),
                ydb_auth_v1_pb2_grpc.AuthServiceStub,
                "Login",
                _wrap_static_credentials_response,
                settings_impl.BaseRequestSettings().with_timeout(self.request_timeout).with_need_rpc_auth(False),
            )
        finally:
            conn.close()
        return {"expires_in": 30 * 60, "access_token": result.token}

    def _update_driver_config(self, driver_config):
        from .driver import DriverConfig

        self.driver_config = DriverConfig(
            endpoint=driver_config.endpoint,
            database=driver_config.database,
            root_certificates=driver_config.root_certificates,
        )


class AnonymousCredentials(Credentials):
    @staticmethod
    def auth_metadata():
        return []


class AuthTokenCredentials(Credentials):
    def __init__(self, token):
        self._token = token

    def auth_metadata(self):
        return [(YDB_AUTH_TICKET_HEADER, self._token)]


class AccessTokenCredentials(Credentials):
    def __init__(self, token):
        self._token = token

    def auth_metadata(self):
        return [(YDB_AUTH_TICKET_HEADER, self._token)]



# -*- coding: utf-8 -*-
# flake8: noqa
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: protos/ydb_operation.proto
# Protobuf Python Version: 6.31.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    31,
    1,
    '',
    'protos/ydb_operation.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import any_pb2 as google_dot_protobuf_dot_any__pb2
from google.protobuf import duration_pb2 as google_dot_protobuf_dot_duration__pb2
from ydb._grpc.v6.protos.annotations import validation_pb2 as protos_dot_annotations_dot_validation__pb2
from ydb._grpc.v6.protos import ydb_common_pb2 as protos_dot_ydb__common__pb2
from ydb._grpc.v6.protos import ydb_issue_message_pb2 as protos_dot_ydb__issue__message__pb2
from ydb._grpc.v6.protos import ydb_status_codes_pb2 as protos_dot_ydb__status__codes__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1aprotos/ydb_operation.proto\x12\x0eYdb.Operations\x1a\x19google/protobuf/any.proto\x1a\x1egoogle/protobuf/duration.proto\x1a#protos/annotations/validation.proto\x1a\x17protos/ydb_common.proto\x1a\x1eprotos/ydb_issue_message.proto\x1a\x1dprotos/ydb_status_codes.proto\"\xb6\x03\n\x0fOperationParams\x12\x45\n\x0eoperation_mode\x18\x01 \x01(\x0e\x32-.Ydb.Operations.OperationParams.OperationMode\x12\x34\n\x11operation_timeout\x18\x02 \x01(\x0b\x32\x19.google.protobuf.Duration\x12/\n\x0c\x63\x61ncel_after\x18\x03 \x01(\x0b\x32\x19.google.protobuf.Duration\x12M\n\x06labels\x18\x04 \x03(\x0b\x32+.Ydb.Operations.OperationParams.LabelsEntryB\x10\xa2\xe6*\x03\x18\x80\x01\xaa\xe6*\x05\n\x03\x18\x80\x01\x12\x31\n\x10report_cost_info\x18\x05 \x01(\x0e\x32\x17.Ydb.FeatureFlag.Status\x1a-\n\x0bLabelsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"D\n\rOperationMode\x12\x1e\n\x1aOPERATION_MODE_UNSPECIFIED\x10\x00\x12\x08\n\x04SYNC\x10\x01\x12\t\n\x05\x41SYNC\x10\x02\"\'\n\x13GetOperationRequest\x12\x10\n\x02id\x18\x01 \x01(\tB\x04\x90\xe6*\x01\"D\n\x14GetOperationResponse\x12,\n\toperation\x18\x01 \x01(\x0b\x32\x19.Ydb.Operations.Operation\"*\n\x16\x43\x61ncelOperationRequest\x12\x10\n\x02id\x18\x01 \x01(\tB\x04\x90\xe6*\x01\"m\n\x17\x43\x61ncelOperationResponse\x12)\n\x06status\x18\x01 \x01(\x0e\x32\x19.Ydb.StatusIds.StatusCode\x12\'\n\x06issues\x18\x02 \x03(\x0b\x32\x17.Ydb.Issue.IssueMessage\"*\n\x16\x46orgetOperationRequest\x12\x10\n\x02id\x18\x01 \x01(\tB\x04\x90\xe6*\x01\"m\n\x17\x46orgetOperationResponse\x12)\n\x06status\x18\x01 \x01(\x0e\x32\x19.Ydb.StatusIds.StatusCode\x12\'\n\x06issues\x18\x02 \x03(\x0b\x32\x17.Ydb.Issue.IssueMessage\"R\n\x15ListOperationsRequest\x12\x12\n\x04kind\x18\x01 \x01(\tB\x04\x90\xe6*\x01\x12\x11\n\tpage_size\x18\x02 \x01(\x04\x12\x12\n\npage_token\x18\x03 \x01(\t\"\xb4\x01\n\x16ListOperationsResponse\x12)\n\x06status\x18\x01 \x01(\x0e\x32\x19.Ydb.StatusIds.StatusCode\x12\'\n\x06issues\x18\x02 \x03(\x0b\x32\x17.Ydb.Issue.IssueMessage\x12-\n\noperations\x18\x03 \x03(\x0b\x32\x19.Ydb.Operations.Operation\x12\x17\n\x0fnext_page_token\x18\x04 \x01(\t\"\xea\x01\n\tOperation\x12\n\n\x02id\x18\x01 \x01(\t\x12\r\n\x05ready\x18\x02 \x01(\x08\x12)\n\x06status\x18\x03 \x01(\x0e\x32\x19.Ydb.StatusIds.StatusCode\x12\'\n\x06issues\x18\x04 \x03(\x0b\x32\x17.Ydb.Issue.IssueMessage\x12$\n\x06result\x18\x05 \x01(\x0b\x32\x14.google.protobuf.Any\x12&\n\x08metadata\x18\x06 \x01(\x0b\x32\x14.google.protobuf.Any\x12 \n\tcost_info\x18\x07 \x01(\x0b\x32\r.Ydb.CostInfoBc\n\x0etech.ydb.protoB\x0fOperationProtosZ=github.com/ydb-platform/ydb-go-genproto/protos/Ydb_Operations\xf8\x01\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'protos.ydb_operation_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\016tech.ydb.protoB\017OperationProtosZ=github.com/ydb-platform/ydb-go-genproto/protos/Ydb_Operations\370\001\001'
  _globals['_OPERATIONPARAMS_LABELSENTRY']._loaded_options = None
  _globals['_OPERATIONPARAMS_LABELSENTRY']._serialized_options = b'8\001'
  _globals['_OPERATIONPARAMS'].fields_by_name['labels']._loaded_options = None
  _globals['_OPERATIONPARAMS'].fields_by_name['labels']._serialized_options = b'\242\346*\003\030\200\001\252\346*\005\n\003\030\200\001'
  _globals['_GETOPERATIONREQUEST'].fields_by_name['id']._loaded_options = None
  _globals['_GETOPERATIONREQUEST'].fields_by_name['id']._serialized_options = b'\220\346*\001'
  _globals['_CANCELOPERATIONREQUEST'].fields_by_name['id']._loaded_options = None
  _globals['_CANCELOPERATIONREQUEST'].fields_by_name['id']._serialized_options = b'\220\346*\001'
  _globals['_FORGETOPERATIONREQUEST'].fields_by_name['id']._loaded_options = None
  _globals['_FORGETOPERATIONREQUEST'].fields_by_name['id']._serialized_options = b'\220\346*\001'
  _globals['_LISTOPERATIONSREQUEST'].fields_by_name['kind']._loaded_options = None
  _globals['_LISTOPERATIONSREQUEST'].fields_by_name['kind']._serialized_options = b'\220\346*\001'
  _globals['_OPERATIONPARAMS']._serialized_start=231
  _globals['_OPERATIONPARAMS']._serialized_end=669
  _globals['_OPERATIONPARAMS_LABELSENTRY']._serialized_start=554
  _globals['_OPERATIONPARAMS_LABELSENTRY']._serialized_end=599
  _globals['_OPERATIONPARAMS_OPERATIONMODE']._serialized_start=601
  _globals['_OPERATIONPARAMS_OPERATIONMODE']._serialized_end=669
  _globals['_GETOPERATIONREQUEST']._serialized_start=671
  _globals['_GETOPERATIONREQUEST']._serialized_end=710
  _globals['_GETOPERATIONRESPONSE']._serialized_start=712
  _globals['_GETOPERATIONRESPONSE']._serialized_end=780
  _globals['_CANCELOPERATIONREQUEST']._serialized_start=782
  _globals['_CANCELOPERATIONREQUEST']._serialized_end=824
  _globals['_CANCELOPERATIONRESPONSE']._serialized_start=826
  _globals['_CANCELOPERATIONRESPONSE']._serialized_end=935
  _globals['_FORGETOPERATIONREQUEST']._serialized_start=937
  _globals['_FORGETOPERATIONREQUEST']._serialized_end=979
  _globals['_FORGETOPERATIONRESPONSE']._serialized_start=981
  _globals['_FORGETOPERATIONRESPONSE']._serialized_end=1090
  _globals['_LISTOPERATIONSREQUEST']._serialized_start=1092
  _globals['_LISTOPERATIONSREQUEST']._serialized_end=1174
  _globals['_LISTOPERATIONSRESPONSE']._serialized_start=1177
  _globals['_LISTOPERATIONSRESPONSE']._serialized_end=1357
  _globals['_OPERATION']._serialized_start=1360
  _globals['_OPERATION']._serialized_end=1594
# @@protoc_insertion_point(module_scope)



from .._topic_common.common import _get_shared_event_loop, CallFromSyncToAsync
from ..aio.coordination.session import CoordinationSession as CoordinationSessionAio
from .semaphore import CoordinationSemaphore


class CoordinationSession:
    def __init__(self, client, path: str, timeout_sec: float = 5):
        self._client = client
        self._path = path
        self._timeout_sec = timeout_sec

        self._caller = CallFromSyncToAsync(_get_shared_event_loop())
        self._closed = False

        async def _make_session() -> CoordinationSessionAio:
            return CoordinationSessionAio(
                client._driver,
                path,
            )

        self._async_session: CoordinationSessionAio = self._caller.safe_call_with_result(
            _make_session(),
            self._timeout_sec,
        )

    def semaphore(self, name: str, limit: int = 1):
        return CoordinationSemaphore(self, name, limit)

    def close(self):
        if self._closed:
            return
        self._caller.safe_call_with_result(
            self._async_session.close(),
            self._timeout_sec,
        )
        self._closed = True

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc, tb):
        self.close()



import pytest
import ydb

from ydb.query.base import QueryStatsMode
from ydb.query.transaction import QueryTxContext
from ydb.query.transaction import QueryTxStateEnum


class TestQueryTransaction:
    def test_tx_begin(self, tx: QueryTxContext):
        assert tx.tx_id is None

        tx.begin()
        assert tx.tx_id is not None

    def test_tx_allow_double_commit(self, tx: QueryTxContext):
        tx.begin()
        tx.commit()
        tx.commit()

    def test_tx_allow_double_rollback(self, tx: QueryTxContext):
        tx.begin()
        tx.rollback()
        tx.rollback()

    def test_tx_commit_before_begin(self, tx: QueryTxContext):
        tx.commit()
        assert tx._tx_state._state == QueryTxStateEnum.COMMITTED

    def test_tx_rollback_before_begin(self, tx: QueryTxContext):
        tx.rollback()
        assert tx._tx_state._state == QueryTxStateEnum.ROLLBACKED

    def test_tx_first_execute_begins_tx(self, tx: QueryTxContext):
        tx.execute("select 1;")
        tx.commit()

    def test_interactive_tx_commit(self, tx: QueryTxContext):
        tx.execute("select 1;", commit_tx=True)
        with pytest.raises(RuntimeError):
            tx.execute("select 1;")

    def test_tx_execute_raises_after_commit(self, tx: QueryTxContext):
        tx.begin()
        tx.commit()
        with pytest.raises(RuntimeError):
            tx.execute("select 1;")

    def test_tx_execute_raises_after_rollback(self, tx: QueryTxContext):
        tx.begin()
        tx.rollback()
        with pytest.raises(RuntimeError):
            tx.execute("select 1;")

    def test_context_manager_rollbacks_tx(self, tx: QueryTxContext):
        with tx:
            tx.begin()

        assert tx._tx_state._state == QueryTxStateEnum.ROLLBACKED

    def test_context_manager_normal_flow(self, tx: QueryTxContext):
        with tx:
            tx.begin()
            tx.execute("select 1;")
            tx.commit()

        assert tx._tx_state._state == QueryTxStateEnum.COMMITTED

    def test_context_manager_does_not_hide_exceptions(self, tx: QueryTxContext):
        class CustomException(Exception):
            pass

        with pytest.raises(CustomException):
            with tx:
                raise CustomException()

    def test_execute_as_context_manager(self, tx: QueryTxContext):
        tx.begin()

        with tx.execute("select 1;") as results:
            res = [result_set for result_set in results]

        assert len(res) == 1

    def test_execute_two_results(self, tx: QueryTxContext):
        tx.begin()
        counter = 0
        res = []

        with tx.execute("select 1; select 2") as results:
            for result_set in results:
                counter += 1
                res.append(list(result_set.rows[0].values()))

        assert res == [[1], [2]]
        assert counter == 2

    def test_tx_identity_before_begin_raises(self, tx: QueryTxContext):
        with pytest.raises(RuntimeError):
            tx._tx_identity()

    def test_tx_identity_after_begin_works(self, tx: QueryTxContext):
        tx.begin()

        identity = tx._tx_identity()

        assert identity.tx_id == tx.tx_id
        assert identity.session_id == tx.session_id

    @pytest.mark.parametrize(
        "stats_mode",
        [
            None,
            QueryStatsMode.UNSPECIFIED,
            QueryStatsMode.NONE,
            QueryStatsMode.BASIC,
            QueryStatsMode.FULL,
            QueryStatsMode.PROFILE,
        ],
    )
    def test_stats_mode(self, tx: QueryTxContext, stats_mode: QueryStatsMode):
        for _ in tx.execute("SELECT 1; SELECT 2; SELECT 3;", commit_tx=True, stats_mode=stats_mode):
            pass

        stats = tx.last_query_stats

        if stats_mode in [None, QueryStatsMode.NONE, QueryStatsMode.UNSPECIFIED]:
            assert stats is None
            return

        assert stats is not None
        assert len(stats.query_phases) > 0

        if stats_mode != QueryStatsMode.BASIC:
            assert len(stats.query_plan) > 0
        else:
            assert stats.query_plan == ""

    def test_rollback_after_tli_aborted_is_safe(self, pool, table_path: str):
        # Given: a row in a table
        table_ref = f"`{table_path}`"
        pool.execute_with_retries(f"UPSERT INTO {table_ref} (id, i64Val) VALUES (1, 0);")

        # When: two concurrent transactions try to modify this row.
        with pool.checkout() as session1, pool.checkout() as session2:
            tx1 = session1.transaction()
            tx2 = session2.transaction()

            with tx1.execute(f"SELECT i64Val FROM {table_ref} WHERE id = 1;") as _:
                pass

            with tx2.execute(f"UPSERT INTO {table_ref} (id, i64Val) VALUES (1, 1);", commit_tx=True) as _:
                pass

            with tx1.execute(f"UPSERT INTO {table_ref} (id, i64Val) VALUES (1, 2);") as _:
                pass
            with pytest.raises(ydb.Aborted):
                tx1.commit()  # receive TLI here

            # Then: rollback (as a handling of Aborted exception) must be successful.
            tx1.rollback()



# -*- coding: utf-8 -*-
from . import issues
from . import _apis


def _forget_operation_request(operation_id):
    request = _apis.ydb_operation.ForgetOperationRequest(id=operation_id)
    return request


def _forget_operation_response(rpc_state, response):  # pylint: disable=W0613
    issues._process_response(response)


def _cancel_operation_request(operation_id):
    request = _apis.ydb_operation.CancelOperationRequest(id=operation_id)
    return request


def _cancel_operation_response(rpc_state, response):  # pylint: disable=W0613
    issues._process_response(response)


def _get_operation_request(self):
    request = _apis.ydb_operation.GetOperationRequest(id=self.id)
    return request


class OperationClient(object):
    def __init__(self, driver):
        self._driver = driver

    def cancel(self, operation_id, settings=None):
        return self._driver(
            _cancel_operation_request(operation_id),
            _apis.OperationService.Stub,
            _apis.OperationService.CancelOperation,
            _cancel_operation_response,
            settings,
        )

    def forget(self, operation_id, settings=None):
        return self._driver(
            _forget_operation_request(operation_id),
            _apis.OperationService.Stub,
            _apis.OperationService.ForgetOperation,
            _forget_operation_response,
            settings,
        )


class Operation(object):
    __slots__ = ("id", "_driver", "self_cls")

    def __init__(self, rpc_state, response, driver=None):  # pylint: disable=W0613
        # implement proper interface a bit later
        issues._process_response(response.operation)
        self.id = response.operation.id
        self._driver = driver
        # self.ready = operation.ready

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return "<Operation %s>" % (self.id,)

    def _ensure_implements(self):
        if self._driver is None:
            raise ValueError("Operation doesn't implement request!")

    def cancel(self, settings=None):
        self._ensure_implements()
        return self._driver(
            _cancel_operation_request(self.id),
            _apis.OperationService.Stub,
            _apis.OperationService.CancelOperation,
            _cancel_operation_response,
            settings,
        )

    def forget(self, settings=None):
        self._ensure_implements()
        return self._driver(
            _forget_operation_request(self.id),
            _apis.OperationService.Stub,
            _apis.OperationService.ForgetOperation,
            _forget_operation_response,
            settings,
        )

    def get(self, settings=None):
        self._ensure_implements()
        return self._driver(
            _get_operation_request(self),
            _apis.OperationService.Stub,
            _apis.OperationService.GetOperation,
            self.__class__,
            settings,
            (self._driver,),
        )



# -*- coding: utf-8 -*-
# flake8: noqa
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: ydb_export_v1.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ydb._grpc.v3.protos import ydb_export_pb2 as protos_dot_ydb__export__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='ydb_export_v1.proto',
  package='Ydb.Export.V1',
  syntax='proto3',
  serialized_options=b'\n\030tech.ydb.proto.export.v1Z5github.com/ydb-platform/ydb-go-genproto/Ydb_Export_V1',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x13ydb_export_v1.proto\x12\rYdb.Export.V1\x1a\x17protos/ydb_export.proto2\xf6\x01\n\rExportService\x12K\n\nExportToYt\x12\x1d.Ydb.Export.ExportToYtRequest\x1a\x1e.Ydb.Export.ExportToYtResponse\x12K\n\nExportToS3\x12\x1d.Ydb.Export.ExportToS3Request\x1a\x1e.Ydb.Export.ExportToS3Response\x12K\n\nExportToFs\x12\x1d.Ydb.Export.ExportToFsRequest\x1a\x1e.Ydb.Export.ExportToFsResponseBQ\n\x18tech.ydb.proto.export.v1Z5github.com/ydb-platform/ydb-go-genproto/Ydb_Export_V1b\x06proto3'
  ,
  dependencies=[protos_dot_ydb__export__pb2.DESCRIPTOR,])



_sym_db.RegisterFileDescriptor(DESCRIPTOR)


DESCRIPTOR._options = None

_EXPORTSERVICE = _descriptor.ServiceDescriptor(
  name='ExportService',
  full_name='Ydb.Export.V1.ExportService',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=64,
  serialized_end=310,
  methods=[
  _descriptor.MethodDescriptor(
    name='ExportToYt',
    full_name='Ydb.Export.V1.ExportService.ExportToYt',
    index=0,
    containing_service=None,
    input_type=protos_dot_ydb__export__pb2._EXPORTTOYTREQUEST,
    output_type=protos_dot_ydb__export__pb2._EXPORTTOYTRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ExportToS3',
    full_name='Ydb.Export.V1.ExportService.ExportToS3',
    index=1,
    containing_service=None,
    input_type=protos_dot_ydb__export__pb2._EXPORTTOS3REQUEST,
    output_type=protos_dot_ydb__export__pb2._EXPORTTOS3RESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ExportToFs',
    full_name='Ydb.Export.V1.ExportService.ExportToFs',
    index=2,
    containing_service=None,
    input_type=protos_dot_ydb__export__pb2._EXPORTTOFSREQUEST,
    output_type=protos_dot_ydb__export__pb2._EXPORTTOFSRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_EXPORTSERVICE)

DESCRIPTOR.services_by_name['ExportService'] = _EXPORTSERVICE

# @@protoc_insertion_point(module_scope)



import asyncio
import json

import pytest
import ydb

from typing import Optional

from ydb import QueryExplainResultFormat
from ydb.aio.query.pool import QuerySessionPool
from ydb.aio.query.session import QuerySession
from ydb.aio.query.transaction import QueryTxContext


class TestQuerySessionPool:
    @pytest.mark.asyncio
    async def test_checkout_provides_created_session(self, pool: QuerySessionPool):
        async with pool.checkout() as session:
            assert session.is_active

    @pytest.mark.asyncio
    async def test_oneshot_query_normal(self, pool: QuerySessionPool):
        res = await pool.execute_with_retries("select 1;")
        assert len(res) == 1

    @pytest.mark.asyncio
    async def test_oneshot_ddl_query(self, pool: QuerySessionPool):
        await pool.execute_with_retries("drop table if exists Queen;")
        await pool.execute_with_retries("create table Queen(key UInt64, PRIMARY KEY (key));")
        await pool.execute_with_retries("drop table Queen;")

    @pytest.mark.asyncio
    async def test_oneshot_query_raises(self, pool: QuerySessionPool):
        with pytest.raises(ydb.GenericError):
            await pool.execute_with_retries("Is this the real life? Is this just fantasy?")

    @pytest.mark.asyncio
    async def test_retry_op_uses_created_session(self, pool: QuerySessionPool):
        async def callee(session: QuerySession):
            assert session.is_active

        await pool.retry_operation_async(callee)

    @pytest.mark.asyncio
    async def test_retry_op_normal(self, pool: QuerySessionPool):
        async def callee(session: QuerySession):
            async with session.transaction() as tx:
                iterator = await tx.execute("select 1;", commit_tx=True)
                return [result_set async for result_set in iterator]

        res = await pool.retry_operation_async(callee)
        assert len(res) == 1

    @pytest.mark.asyncio
    async def test_retry_op_raises(self, pool: QuerySessionPool):
        class CustomException(Exception):
            pass

        async def callee(session: QuerySession):
            raise CustomException()

        with pytest.raises(CustomException):
            await pool.retry_operation_async(callee)

    @pytest.mark.parametrize(
        "tx_mode",
        [
            (None),
            (ydb.QuerySerializableReadWrite()),
            (ydb.QuerySnapshotReadOnly()),
            (ydb.QuerySnapshotReadWrite()),
            (ydb.QueryOnlineReadOnly()),
            (ydb.QueryStaleReadOnly()),
        ],
    )
    @pytest.mark.asyncio
    async def test_retry_tx_normal(self, pool: QuerySessionPool, tx_mode: Optional[ydb.BaseQueryTxMode]):
        retry_no = 0

        async def callee(tx: QueryTxContext):
            nonlocal retry_no
            if retry_no < 2:
                retry_no += 1
                raise ydb.Unavailable("Fake fast backoff error")
            result_stream = await tx.execute("SELECT 1")
            return [result_set async for result_set in result_stream]

        result = await pool.retry_tx_async(callee=callee, tx_mode=tx_mode)
        assert len(result) == 1
        assert retry_no == 2

    @pytest.mark.asyncio
    async def test_retry_tx_raises(self, pool: QuerySessionPool):
        class CustomException(Exception):
            pass

        async def callee(tx: QueryTxContext):
            raise CustomException()

        with pytest.raises(CustomException):
            await pool.retry_tx_async(callee)

    @pytest.mark.asyncio
    async def test_pool_size_limit_logic(self, pool: QuerySessionPool):
        target_size = 5
        pool._size = target_size
        ids = set()

        for i in range(1, target_size + 1):
            session = await pool.acquire()
            assert pool._current_size == i
            assert session.session_id not in ids
            ids.add(session.session_id)

        with pytest.raises(asyncio.TimeoutError):
            await asyncio.wait_for(pool.acquire(), timeout=0.1)

        last_id = session.session_id
        await pool.release(session)

        session = await pool.acquire()
        assert session.session_id == last_id
        assert pool._current_size == target_size

    @pytest.mark.asyncio
    async def test_checkout_do_not_increase_size(self, pool: QuerySessionPool):
        session_id = None
        for _ in range(10):
            async with pool.checkout() as session:
                if session_id is None:
                    session_id = session.session_id
                assert pool._current_size == 1
                assert session_id == session.session_id

    @pytest.mark.asyncio
    async def test_pool_recreates_bad_sessions(self, pool: QuerySessionPool):
        async with pool.checkout() as session:
            session_id = session.session_id
            await session.delete()

        async with pool.checkout() as session:
            assert session_id != session.session_id
            assert pool._current_size == 1

    @pytest.mark.asyncio
    async def test_acquire_from_closed_pool_raises(self, pool: QuerySessionPool):
        await pool.stop()
        with pytest.raises(RuntimeError):
            await pool.acquire()

    @pytest.mark.asyncio
    async def test_acquire_with_timeout_from_closed_pool_raises(self, pool: QuerySessionPool):
        await pool.stop()
        with pytest.raises(RuntimeError):
            await asyncio.wait_for(pool.acquire(), timeout=0.1)

    @pytest.mark.asyncio
    async def test_no_session_leak(self, driver, docker_project):
        pool = ydb.aio.QuerySessionPool(driver, 1)
        docker_project.stop()
        try:
            await asyncio.wait_for(pool.acquire(), timeout=0.1)
        except ydb.Error:
            pass
        assert pool._current_size == 0

        docker_project.start()
        await pool.stop()

    @pytest.mark.asyncio
    async def test_acquire_no_race_condition(self, driver):
        ids = set()
        async with ydb.aio.QuerySessionPool(driver, 1) as pool:

            async def acquire_session():
                session = await pool.acquire()
                ids.add(session.session_id)
                await pool.release(session)

            tasks = [acquire_session() for _ in range(10)]

            await asyncio.gather(*tasks)

            assert len(ids) == 1
            assert pool._current_size == 1

    @pytest.mark.asyncio
    async def test_released_after_future_canceled(self, driver):
        async with ydb.aio.QuerySessionPool(driver, 1) as pool:
            s = await pool.acquire()
            waiter = asyncio.ensure_future(pool.acquire())
            await asyncio.sleep(0)  # let waiter run until waiting point
            waiter.cancel()
            await pool.release(s)
            await asyncio.wait([waiter])

            assert pool._queue.qsize() == 1

    @pytest.mark.asyncio
    async def test_explain_with_retries(self, pool: QuerySessionPool):
        await pool.execute_with_retries("DROP TABLE IF EXISTS test_explain")
        await pool.execute_with_retries("CREATE TABLE test_explain (id Int64, PRIMARY KEY (id))")
        try:
            plan = await pool.explain_with_retries(
                "SELECT * FROM test_explain", result_format=QueryExplainResultFormat.STR
            )
            isinstance(plan, str)
            assert "FullScan" in plan

            plan = await pool.explain_with_retries(
                "SELECT * FROM test_explain", result_format=QueryExplainResultFormat.DICT
            )
            plan_string = json.dumps(plan)
            assert "FullScan" in plan_string

            plan = await pool.explain_with_retries(
                "SELECT * FROM test_explain WHERE id = $id",
                {"$id": 1},
                result_format=QueryExplainResultFormat.DICT,
            )
            plan_string = json.dumps(plan)
            assert "Lookup" in plan_string
        finally:
            await pool.execute_with_retries("DROP TABLE test_explain")



# -*- coding: utf-8 -*-
# flake8: noqa
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: ydb_operation_v1.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ydb._grpc.v3.protos import ydb_operation_pb2 as protos_dot_ydb__operation__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='ydb_operation_v1.proto',
  package='Ydb.Operation.V1',
  syntax='proto3',
  serialized_options=b'\n\033tech.ydb.proto.operation.v1Z8github.com/ydb-platform/ydb-go-genproto/Ydb_Operation_V1',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x16ydb_operation_v1.proto\x12\x10Ydb.Operation.V1\x1a\x1aprotos/ydb_operation.proto2\x96\x03\n\x10OperationService\x12Y\n\x0cGetOperation\x12#.Ydb.Operations.GetOperationRequest\x1a$.Ydb.Operations.GetOperationResponse\x12\x62\n\x0f\x43\x61ncelOperation\x12&.Ydb.Operations.CancelOperationRequest\x1a\'.Ydb.Operations.CancelOperationResponse\x12\x62\n\x0f\x46orgetOperation\x12&.Ydb.Operations.ForgetOperationRequest\x1a\'.Ydb.Operations.ForgetOperationResponse\x12_\n\x0eListOperations\x12%.Ydb.Operations.ListOperationsRequest\x1a&.Ydb.Operations.ListOperationsResponseBW\n\x1btech.ydb.proto.operation.v1Z8github.com/ydb-platform/ydb-go-genproto/Ydb_Operation_V1b\x06proto3'
  ,
  dependencies=[protos_dot_ydb__operation__pb2.DESCRIPTOR,])



_sym_db.RegisterFileDescriptor(DESCRIPTOR)


DESCRIPTOR._options = None

_OPERATIONSERVICE = _descriptor.ServiceDescriptor(
  name='OperationService',
  full_name='Ydb.Operation.V1.OperationService',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=73,
  serialized_end=479,
  methods=[
  _descriptor.MethodDescriptor(
    name='GetOperation',
    full_name='Ydb.Operation.V1.OperationService.GetOperation',
    index=0,
    containing_service=None,
    input_type=protos_dot_ydb__operation__pb2._GETOPERATIONREQUEST,
    output_type=protos_dot_ydb__operation__pb2._GETOPERATIONRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='CancelOperation',
    full_name='Ydb.Operation.V1.OperationService.CancelOperation',
    index=1,
    containing_service=None,
    input_type=protos_dot_ydb__operation__pb2._CANCELOPERATIONREQUEST,
    output_type=protos_dot_ydb__operation__pb2._CANCELOPERATIONRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ForgetOperation',
    full_name='Ydb.Operation.V1.OperationService.ForgetOperation',
    index=2,
    containing_service=None,
    input_type=protos_dot_ydb__operation__pb2._FORGETOPERATIONREQUEST,
    output_type=protos_dot_ydb__operation__pb2._FORGETOPERATIONRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ListOperations',
    full_name='Ydb.Operation.V1.OperationService.ListOperations',
    index=3,
    containing_service=None,
    input_type=protos_dot_ydb__operation__pb2._LISTOPERATIONSREQUEST,
    output_type=protos_dot_ydb__operation__pb2._LISTOPERATIONSRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_OPERATIONSERVICE)

DESCRIPTOR.services_by_name['OperationService'] = _OPERATIONSERVICE

# @@protoc_insertion_point(module_scope)



from protos import ydb_operation_pb2 as _ydb_operation_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from collections.abc import Iterable as _Iterable, Mapping as _Mapping
from typing import ClassVar as _ClassVar, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class StorageUnits(_message.Message):
    __slots__ = ("unit_kind", "count")
    UNIT_KIND_FIELD_NUMBER: _ClassVar[int]
    COUNT_FIELD_NUMBER: _ClassVar[int]
    unit_kind: str
    count: int
    def __init__(self, unit_kind: _Optional[str] = ..., count: _Optional[int] = ...) -> None: ...

class ComputationalUnits(_message.Message):
    __slots__ = ("unit_kind", "availability_zone", "count")
    UNIT_KIND_FIELD_NUMBER: _ClassVar[int]
    AVAILABILITY_ZONE_FIELD_NUMBER: _ClassVar[int]
    COUNT_FIELD_NUMBER: _ClassVar[int]
    unit_kind: str
    availability_zone: str
    count: int
    def __init__(self, unit_kind: _Optional[str] = ..., availability_zone: _Optional[str] = ..., count: _Optional[int] = ...) -> None: ...

class AllocatedComputationalUnit(_message.Message):
    __slots__ = ("host", "port", "unit_kind")
    HOST_FIELD_NUMBER: _ClassVar[int]
    PORT_FIELD_NUMBER: _ClassVar[int]
    UNIT_KIND_FIELD_NUMBER: _ClassVar[int]
    host: str
    port: int
    unit_kind: str
    def __init__(self, host: _Optional[str] = ..., port: _Optional[int] = ..., unit_kind: _Optional[str] = ...) -> None: ...

class Resources(_message.Message):
    __slots__ = ("storage_units", "computational_units")
    STORAGE_UNITS_FIELD_NUMBER: _ClassVar[int]
    COMPUTATIONAL_UNITS_FIELD_NUMBER: _ClassVar[int]
    storage_units: _containers.RepeatedCompositeFieldContainer[StorageUnits]
    computational_units: _containers.RepeatedCompositeFieldContainer[ComputationalUnits]
    def __init__(self, storage_units: _Optional[_Iterable[_Union[StorageUnits, _Mapping]]] = ..., computational_units: _Optional[_Iterable[_Union[ComputationalUnits, _Mapping]]] = ...) -> None: ...

class ServerlessResources(_message.Message):
    __slots__ = ("shared_database_path",)
    SHARED_DATABASE_PATH_FIELD_NUMBER: _ClassVar[int]
    shared_database_path: str
    def __init__(self, shared_database_path: _Optional[str] = ...) -> None: ...

class DatabaseOptions(_message.Message):
    __slots__ = ("disable_tx_service", "disable_external_subdomain", "plan_resolution")
    DISABLE_TX_SERVICE_FIELD_NUMBER: _ClassVar[int]
    DISABLE_EXTERNAL_SUBDOMAIN_FIELD_NUMBER: _ClassVar[int]
    PLAN_RESOLUTION_FIELD_NUMBER: _ClassVar[int]
    disable_tx_service: bool
    disable_external_subdomain: bool
    plan_resolution: int
    def __init__(self, disable_tx_service: bool = ..., disable_external_subdomain: bool = ..., plan_resolution: _Optional[int] = ...) -> None: ...

class SchemaOperationQuotas(_message.Message):
    __slots__ = ("leaky_bucket_quotas",)
    class LeakyBucket(_message.Message):
        __slots__ = ("bucket_size", "bucket_seconds")
        BUCKET_SIZE_FIELD_NUMBER: _ClassVar[int]
        BUCKET_SECONDS_FIELD_NUMBER: _ClassVar[int]
        bucket_size: float
        bucket_seconds: int
        def __init__(self, bucket_size: _Optional[float] = ..., bucket_seconds: _Optional[int] = ...) -> None: ...
    LEAKY_BUCKET_QUOTAS_FIELD_NUMBER: _ClassVar[int]
    leaky_bucket_quotas: _containers.RepeatedCompositeFieldContainer[SchemaOperationQuotas.LeakyBucket]
    def __init__(self, leaky_bucket_quotas: _Optional[_Iterable[_Union[SchemaOperationQuotas.LeakyBucket, _Mapping]]] = ...) -> None: ...

class DatabaseQuotas(_message.Message):
    __slots__ = ("data_size_hard_quota", "data_size_soft_quota", "data_stream_shards_quota", "data_stream_reserved_storage_quota", "ttl_min_run_internal_seconds")
    DATA_SIZE_HARD_QUOTA_FIELD_NUMBER: _ClassVar[int]
    DATA_SIZE_SOFT_QUOTA_FIELD_NUMBER: _ClassVar[int]
    DATA_STREAM_SHARDS_QUOTA_FIELD_NUMBER: _ClassVar[int]
    DATA_STREAM_RESERVED_STORAGE_QUOTA_FIELD_NUMBER: _ClassVar[int]
    TTL_MIN_RUN_INTERNAL_SECONDS_FIELD_NUMBER: _ClassVar[int]
    data_size_hard_quota: int
    data_size_soft_quota: int
    data_stream_shards_quota: int
    data_stream_reserved_storage_quota: int
    ttl_min_run_internal_seconds: int
    def __init__(self, data_size_hard_quota: _Optional[int] = ..., data_size_soft_quota: _Optional[int] = ..., data_stream_shards_quota: _Optional[int] = ..., data_stream_reserved_storage_quota: _Optional[int] = ..., ttl_min_run_internal_seconds: _Optional[int] = ...) -> None: ...

class CreateDatabaseRequest(_message.Message):
    __slots__ = ("operation_params", "path", "resources", "shared_resources", "serverless_resources", "options", "attributes", "schema_operation_quotas", "idempotency_key", "database_quotas")
    class AttributesEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
    OPERATION_PARAMS_FIELD_NUMBER: _ClassVar[int]
    PATH_FIELD_NUMBER: _ClassVar[int]
    RESOURCES_FIELD_NUMBER: _ClassVar[int]
    SHARED_RESOURCES_FIELD_NUMBER: _ClassVar[int]
    SERVERLESS_RESOURCES_FIELD_NUMBER: _ClassVar[int]
    OPTIONS_FIELD_NUMBER: _ClassVar[int]
    ATTRIBUTES_FIELD_NUMBER: _ClassVar[int]
    SCHEMA_OPERATION_QUOTAS_FIELD_NUMBER: _ClassVar[int]
    IDEMPOTENCY_KEY_FIELD_NUMBER: _ClassVar[int]
    DATABASE_QUOTAS_FIELD_NUMBER: _ClassVar[int]
    operation_params: _ydb_operation_pb2.OperationParams
    path: str
    resources: Resources
    shared_resources: Resources
    serverless_resources: ServerlessResources
    options: DatabaseOptions
    attributes: _containers.ScalarMap[str, str]
    schema_operation_quotas: SchemaOperationQuotas
    idempotency_key: str
    database_quotas: DatabaseQuotas
    def __init__(self, operation_params: _Optional[_Union[_ydb_operation_pb2.OperationParams, _Mapping]] = ..., path: _Optional[str] = ..., resources: _Optional[_Union[Resources, _Mapping]] = ..., shared_resources: _Optional[_Union[Resources, _Mapping]] = ..., serverless_resources: _Optional[_Union[ServerlessResources, _Mapping]] = ..., options: _Optional[_Union[DatabaseOptions, _Mapping]] = ..., attributes: _Optional[_Mapping[str, str]] = ..., schema_operation_quotas: _Optional[_Union[SchemaOperationQuotas, _Mapping]] = ..., idempotency_key: _Optional[str] = ..., database_quotas: _Optional[_Union[DatabaseQuotas, _Mapping]] = ...) -> None: ...

class CreateDatabaseResponse(_message.Message):
    __slots__ = ("operation",)
    OPERATION_FIELD_NUMBER: _ClassVar[int]
    operation: _ydb_operation_pb2.Operation
    def __init__(self, operation: _Optional[_Union[_ydb_operation_pb2.Operation, _Mapping]] = ...) -> None: ...

class GetDatabaseStatusRequest(_message.Message):
    __slots__ = ("path", "operation_params")
    PATH_FIELD_NUMBER: _ClassVar[int]
    OPERATION_PARAMS_FIELD_NUMBER: _ClassVar[int]
    path: str
    operation_params: _ydb_operation_pb2.OperationParams
    def __init__(self, path: _Optional[str] = ..., operation_params: _Optional[_Union[_ydb_operation_pb2.OperationParams, _Mapping]] = ...) -> None: ...

class GetDatabaseStatusResponse(_message.Message):
    __slots__ = ("operation",)
    OPERATION_FIELD_NUMBER: _ClassVar[int]
    operation: _ydb_operation_pb2.Operation
    def __init__(self, operation: _Optional[_Union[_ydb_operation_pb2.Operation, _Mapping]] = ...) -> None: ...

class GetDatabaseStatusResult(_message.Message):
    __slots__ = ("path", "state", "required_resources", "required_shared_resources", "serverless_resources", "allocated_resources", "registered_resources", "generation", "schema_operation_quotas", "database_quotas")
    class State(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        STATE_UNSPECIFIED: _ClassVar[GetDatabaseStatusResult.State]
        CREATING: _ClassVar[GetDatabaseStatusResult.State]
        RUNNING: _ClassVar[GetDatabaseStatusResult.State]
        REMOVING: _ClassVar[GetDatabaseStatusResult.State]
        PENDING_RESOURCES: _ClassVar[GetDatabaseStatusResult.State]
        CONFIGURING: _ClassVar[GetDatabaseStatusResult.State]
    STATE_UNSPECIFIED: GetDatabaseStatusResult.State
    CREATING: GetDatabaseStatusResult.State
    RUNNING: GetDatabaseStatusResult.State
    REMOVING: GetDatabaseStatusResult.State
    PENDING_RESOURCES: GetDatabaseStatusResult.State
    CONFIGURING: GetDatabaseStatusResult.State
    PATH_FIELD_NUMBER: _ClassVar[int]
    STATE_FIELD_NUMBER: _ClassVar[int]
    REQUIRED_RESOURCES_FIELD_NUMBER: _ClassVar[int]
    REQUIRED_SHARED_RESOURCES_FIELD_NUMBER: _ClassVar[int]
    SERVERLESS_RESOURCES_FIELD_NUMBER: _ClassVar[int]
    ALLOCATED_RESOURCES_FIELD_NUMBER: _ClassVar[int]
    REGISTERED_RESOURCES_FIELD_NUMBER: _ClassVar[int]
    GENERATION_FIELD_NUMBER: _ClassVar[int]
    SCHEMA_OPERATION_QUOTAS_FIELD_NUMBER: _ClassVar[int]
    DATABASE_QUOTAS_FIELD_NUMBER: _ClassVar[int]
    path: str
    state: GetDatabaseStatusResult.State
    required_resources: Resources
    required_shared_resources: Resources
    serverless_resources: ServerlessResources
    allocated_resources: Resources
    registered_resources: _containers.RepeatedCompositeFieldContainer[AllocatedComputationalUnit]
    generation: int
    schema_operation_quotas: SchemaOperationQuotas
    database_quotas: DatabaseQuotas
    def __init__(self, path: _Optional[str] = ..., state: _Optional[_Union[GetDatabaseStatusResult.State, str]] = ..., required_resources: _Optional[_Union[Resources, _Mapping]] = ..., required_shared_resources: _Optional[_Union[Resources, _Mapping]] = ..., serverless_resources: _Optional[_Union[ServerlessResources, _Mapping]] = ..., allocated_resources: _Optional[_Union[Resources, _Mapping]] = ..., registered_resources: _Optional[_Iterable[_Union[AllocatedComputationalUnit, _Mapping]]] = ..., generation: _Optional[int] = ..., schema_operation_quotas: _Optional[_Union[SchemaOperationQuotas, _Mapping]] = ..., database_quotas: _Optional[_Union[DatabaseQuotas, _Mapping]] = ...) -> None: ...

class AlterDatabaseRequest(_message.Message):
    __slots__ = ("path", "computational_units_to_add", "computational_units_to_remove", "storage_units_to_add", "computational_units_to_register", "computational_units_to_deregister", "operation_params", "generation", "schema_operation_quotas", "idempotency_key", "database_quotas", "alter_attributes")
    class AlterAttributesEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
    PATH_FIELD_NUMBER: _ClassVar[int]
    COMPUTATIONAL_UNITS_TO_ADD_FIELD_NUMBER: _ClassVar[int]
    COMPUTATIONAL_UNITS_TO_REMOVE_FIELD_NUMBER: _ClassVar[int]
    STORAGE_UNITS_TO_ADD_FIELD_NUMBER: _ClassVar[int]
    COMPUTATIONAL_UNITS_TO_REGISTER_FIELD_NUMBER: _ClassVar[int]
    COMPUTATIONAL_UNITS_TO_DEREGISTER_FIELD_NUMBER: _ClassVar[int]
    OPERATION_PARAMS_FIELD_NUMBER: _ClassVar[int]
    GENERATION_FIELD_NUMBER: _ClassVar[int]
    SCHEMA_OPERATION_QUOTAS_FIELD_NUMBER: _ClassVar[int]
    IDEMPOTENCY_KEY_FIELD_NUMBER: _ClassVar[int]
    DATABASE_QUOTAS_FIELD_NUMBER: _ClassVar[int]
    ALTER_ATTRIBUTES_FIELD_NUMBER: _ClassVar[int]
    path: str
    computational_units_to_add: _containers.RepeatedCompositeFieldContainer[ComputationalUnits]
    computational_units_to_remove: _containers.RepeatedCompositeFieldContainer[ComputationalUnits]
    storage_units_to_add: _containers.RepeatedCompositeFieldContainer[StorageUnits]
    computational_units_to_register: _containers.RepeatedCompositeFieldContainer[AllocatedComputationalUnit]
    computational_units_to_deregister: _containers.RepeatedCompositeFieldContainer[AllocatedComputationalUnit]
    operation_params: _ydb_operation_pb2.OperationParams
    generation: int
    schema_operation_quotas: SchemaOperationQuotas
    idempotency_key: str
    database_quotas: DatabaseQuotas
    alter_attributes: _containers.ScalarMap[str, str]
    def __init__(self, path: _Optional[str] = ..., computational_units_to_add: _Optional[_Iterable[_Union[ComputationalUnits, _Mapping]]] = ..., computational_units_to_remove: _Optional[_Iterable[_Union[ComputationalUnits, _Mapping]]] = ..., storage_units_to_add: _Optional[_Iterable[_Union[StorageUnits, _Mapping]]] = ..., computational_units_to_register: _Optional[_Iterable[_Union[AllocatedComputationalUnit, _Mapping]]] = ..., computational_units_to_deregister: _Optional[_Iterable[_Union[AllocatedComputationalUnit, _Mapping]]] = ..., operation_params: _Optional[_Union[_ydb_operation_pb2.OperationParams, _Mapping]] = ..., generation: _Optional[int] = ..., schema_operation_quotas: _Optional[_Union[SchemaOperationQuotas, _Mapping]] = ..., idempotency_key: _Optional[str] = ..., database_quotas: _Optional[_Union[DatabaseQuotas, _Mapping]] = ..., alter_attributes: _Optional[_Mapping[str, str]] = ...) -> None: ...

class AlterDatabaseResponse(_message.Message):
    __slots__ = ("operation",)
    OPERATION_FIELD_NUMBER: _ClassVar[int]
    operation: _ydb_operation_pb2.Operation
    def __init__(self, operation: _Optional[_Union[_ydb_operation_pb2.Operation, _Mapping]] = ...) -> None: ...

class ListDatabasesRequest(_message.Message):
    __slots__ = ("operation_params",)
    OPERATION_PARAMS_FIELD_NUMBER: _ClassVar[int]
    operation_params: _ydb_operation_pb2.OperationParams
    def __init__(self, operation_params: _Optional[_Union[_ydb_operation_pb2.OperationParams, _Mapping]] = ...) -> None: ...

class ListDatabasesResponse(_message.Message):
    __slots__ = ("operation",)
    OPERATION_FIELD_NUMBER: _ClassVar[int]
    operation: _ydb_operation_pb2.Operation
    def __init__(self, operation: _Optional[_Union[_ydb_operation_pb2.Operation, _Mapping]] = ...) -> None: ...

class ListDatabasesResult(_message.Message):
    __slots__ = ("paths",)
    PATHS_FIELD_NUMBER: _ClassVar[int]
    paths: _containers.RepeatedScalarFieldContainer[str]
    def __init__(self, paths: _Optional[_Iterable[str]] = ...) -> None: ...

class RemoveDatabaseRequest(_message.Message):
    __slots__ = ("path", "operation_params")
    PATH_FIELD_NUMBER: _ClassVar[int]
    OPERATION_PARAMS_FIELD_NUMBER: _ClassVar[int]
    path: str
    operation_params: _ydb_operation_pb2.OperationParams
    def __init__(self, path: _Optional[str] = ..., operation_params: _Optional[_Union[_ydb_operation_pb2.OperationParams, _Mapping]] = ...) -> None: ...

class RemoveDatabaseResponse(_message.Message):
    __slots__ = ("operation",)
    OPERATION_FIELD_NUMBER: _ClassVar[int]
    operation: _ydb_operation_pb2.Operation
    def __init__(self, operation: _Optional[_Union[_ydb_operation_pb2.Operation, _Mapping]] = ...) -> None: ...

class StorageUnitDescription(_message.Message):
    __slots__ = ("kind", "labels")
    class LabelsEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
    KIND_FIELD_NUMBER: _ClassVar[int]
    LABELS_FIELD_NUMBER: _ClassVar[int]
    kind: str
    labels: _containers.ScalarMap[str, str]
    def __init__(self, kind: _Optional[str] = ..., labels: _Optional[_Mapping[str, str]] = ...) -> None: ...

class AvailabilityZoneDescription(_message.Message):
    __slots__ = ("name", "labels")
    class LabelsEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
    NAME_FIELD_NUMBER: _ClassVar[int]
    LABELS_FIELD_NUMBER: _ClassVar[int]
    name: str
    labels: _containers.ScalarMap[str, str]
    def __init__(self, name: _Optional[str] = ..., labels: _Optional[_Mapping[str, str]] = ...) -> None: ...

class ComputationalUnitDescription(_message.Message):
    __slots__ = ("kind", "labels", "allowed_availability_zones")
    class LabelsEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
    KIND_FIELD_NUMBER: _ClassVar[int]
    LABELS_FIELD_NUMBER: _ClassVar[int]
    ALLOWED_AVAILABILITY_ZONES_FIELD_NUMBER: _ClassVar[int]
    kind: str
    labels: _containers.ScalarMap[str, str]
    allowed_availability_zones: _containers.RepeatedScalarFieldContainer[str]
    def __init__(self, kind: _Optional[str] = ..., labels: _Optional[_Mapping[str, str]] = ..., allowed_availability_zones: _Optional[_Iterable[str]] = ...) -> None: ...

class DescribeDatabaseOptionsRequest(_message.Message):
    __slots__ = ("operation_params",)
    OPERATION_PARAMS_FIELD_NUMBER: _ClassVar[int]
    operation_params: _ydb_operation_pb2.OperationParams
    def __init__(self, operation_params: _Optional[_Union[_ydb_operation_pb2.OperationParams, _Mapping]] = ...) -> None: ...

class DescribeDatabaseOptionsResponse(_message.Message):
    __slots__ = ("operation",)
    OPERATION_FIELD_NUMBER: _ClassVar[int]
    operation: _ydb_operation_pb2.Operation
    def __init__(self, operation: _Optional[_Union[_ydb_operation_pb2.Operation, _Mapping]] = ...) -> None: ...

class DescribeDatabaseOptionsResult(_message.Message):
    __slots__ = ("storage_units", "availability_zones", "computational_units")
    STORAGE_UNITS_FIELD_NUMBER: _ClassVar[int]
    AVAILABILITY_ZONES_FIELD_NUMBER: _ClassVar[int]
    COMPUTATIONAL_UNITS_FIELD_NUMBER: _ClassVar[int]
    storage_units: _containers.RepeatedCompositeFieldContainer[StorageUnitDescription]
    availability_zones: _containers.RepeatedCompositeFieldContainer[AvailabilityZoneDescription]
    computational_units: _containers.RepeatedCompositeFieldContainer[ComputationalUnitDescription]
    def __init__(self, storage_units: _Optional[_Iterable[_Union[StorageUnitDescription, _Mapping]]] = ..., availability_zones: _Optional[_Iterable[_Union[AvailabilityZoneDescription, _Mapping]]] = ..., computational_units: _Optional[_Iterable[_Union[ComputationalUnitDescription, _Mapping]]] = ...) -> None: ...



# -*- coding: utf-8 -*-
# flake8: noqa
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: protos/ydb_federation_discovery.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ydb._grpc.v4.protos import ydb_operation_pb2 as protos_dot_ydb__operation__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n%protos/ydb_federation_discovery.proto\x12\x17Ydb.FederationDiscovery\x1a\x1aprotos/ydb_operation.proto\"\xf9\x01\n\x0c\x44\x61tabaseInfo\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0c\n\x04path\x18\x02 \x01(\t\x12\n\n\x02id\x18\x03 \x01(\t\x12\x10\n\x08\x65ndpoint\x18\x04 \x01(\t\x12\x10\n\x08location\x18\x05 \x01(\t\x12<\n\x06status\x18\x06 \x01(\x0e\x32,.Ydb.FederationDiscovery.DatabaseInfo.Status\x12\x0e\n\x06weight\x18\x07 \x01(\x03\"O\n\x06Status\x12\x16\n\x12STATUS_UNSPECIFIED\x10\x00\x12\r\n\tAVAILABLE\x10\x01\x12\r\n\tREAD_ONLY\x10\x02\x12\x0f\n\x0bUNAVAILABLE\x10\x03\" \n\x1eListFederationDatabasesRequest\"O\n\x1fListFederationDatabasesResponse\x12,\n\toperation\x18\x01 \x01(\x0b\x32\x19.Ydb.Operations.Operation\"\x9b\x01\n\x1dListFederationDatabasesResult\x12\x1e\n\x16\x63ontrol_plane_endpoint\x18\x01 \x01(\t\x12\x43\n\x14\x66\x65\x64\x65ration_databases\x18\x02 \x03(\x0b\x32%.Ydb.FederationDiscovery.DatabaseInfo\x12\x15\n\rself_location\x18\x03 \x01(\tB\x8b\x01\n#tech.ydb.proto.federation.discoveryB\x19\x46\x65\x64\x65rationDiscoveryProtosZFgithub.com/ydb-platform/ydb-go-genproto/protos/Ydb_FederationDiscovery\xf8\x01\x01\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'protos.ydb_federation_discovery_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n#tech.ydb.proto.federation.discoveryB\031FederationDiscoveryProtosZFgithub.com/ydb-platform/ydb-go-genproto/protos/Ydb_FederationDiscovery\370\001\001'
  _DATABASEINFO._serialized_start=95
  _DATABASEINFO._serialized_end=344
  _DATABASEINFO_STATUS._serialized_start=265
  _DATABASEINFO_STATUS._serialized_end=344
  _LISTFEDERATIONDATABASESREQUEST._serialized_start=346
  _LISTFEDERATIONDATABASESREQUEST._serialized_end=378
  _LISTFEDERATIONDATABASESRESPONSE._serialized_start=380
  _LISTFEDERATIONDATABASESRESPONSE._serialized_end=459
  _LISTFEDERATIONDATABASESRESULT._serialized_start=462
  _LISTFEDERATIONDATABASESRESULT._serialized_end=617
# @@protoc_insertion_point(module_scope)



import asyncio

import pytest
import ydb


@pytest.mark.asyncio
async def test_simple_acquire(driver):
    pool = ydb.aio.SessionPool(driver, 1)
    session = await pool.acquire()

    for _ in range(10):
        with pytest.raises(ydb.SessionPoolEmpty):
            await pool.acquire(timeout=0.1)
    assert session.initialized()

    await pool.release(session)
    await pool.acquire()
    assert session.initialized()
    await pool.stop(timeout=10)


@pytest.mark.asyncio
async def test_waiter_is_notified(driver):
    pool = ydb.aio.SessionPool(driver, 1)
    session = await pool.acquire()
    task = asyncio.ensure_future(pool.acquire())
    await pool.release(session)
    sess = await asyncio.wait_for(task, timeout=0.1)
    assert sess.initialized()
    await pool.stop()


@pytest.mark.asyncio
async def test_no_race_after_future_cancel(driver):
    pool = ydb.aio.SessionPool(driver, 1)

    s = await pool.acquire()
    waiter = asyncio.ensure_future(pool.acquire())
    await asyncio.sleep(0)
    waiter.cancel()
    await pool.release(s)
    await asyncio.wait([waiter])

    assert pool._active_queue.qsize() == 1
    s = await pool.acquire()
    assert s.initialized()
    await pool.stop()


@pytest.mark.asyncio
async def test_release_logic(driver):
    pool = ydb.aio.SessionPool(driver, 1)
    session = await pool.acquire()
    waiter = asyncio.ensure_future(pool.acquire())
    session.reset()
    await pool.release(session)
    res = await waiter
    assert res.initialized()
    await pool.stop()


@pytest.mark.asyncio
async def test_close_basic_logic_case_1(driver):
    pool = ydb.aio.SessionPool(driver, 1)
    s = await pool.acquire()
    waiter = asyncio.ensure_future(pool.acquire())

    await pool.stop()

    with pytest.raises(ValueError):
        waiter.result()

    with pytest.raises(ValueError):
        await pool.acquire()

    await pool.release(s)
    assert pool._active_count == 0


@pytest.mark.asyncio
async def test_no_cluster_endpoints_no_failure(driver, docker_project):
    pool = ydb.aio.SessionPool(driver, 1)
    docker_project.stop()
    waiter = asyncio.ensure_future(pool.acquire())
    docker_project.start()
    sess = await asyncio.wait_for(waiter, 10)
    assert sess.initialized()
    sess.reset()
    await pool.release(sess)
    assert pool._active_count == 0
    await pool.stop()


@pytest.mark.asyncio
async def test_close_basic_logic_case_2(driver):
    pool = ydb.aio.SessionPool(driver, 10)
    acquired = []

    for _ in range(10):
        acquired.append(await pool.acquire())

    for _ in range(3):
        await pool.release(acquired.pop(-1))

    await pool.stop()
    assert pool._active_count == 7

    while acquired:
        await pool.release(acquired.pop(-1))

    assert pool._active_count == 0

    with pytest.raises(ValueError):
        await pool.acquire()

    await pool.stop()


@pytest.mark.asyncio
async def test_min_size_feature(driver):
    pool = ydb.aio.SessionPool(driver, 10, min_pool_size=10)

    await pool.wait_until_min_size()

    assert pool._active_count == 10
    session = await pool.acquire()
    session.reset()
    await pool.release(session)
    await pool.wait_until_min_size()

    await pool.stop()


@pytest.mark.asyncio
async def test_no_session_leak(driver, docker_project):
    pool = ydb.aio.SessionPool(driver, 1)
    docker_project.stop()
    try:
        await pool.acquire(timeout=0.5)
    except ydb.Error:
        pass
    assert pool._active_count == 0

    docker_project.start()
    await pool.stop()
